#!/usr/bin/python
# import networkx as nkx
import os, os.path, sys, re, copy, glob, itertools, datetime, math, random, numpy
import scipy.spatial.distance
import networkx as nkx
from sklearn import decomposition
from scipy.sparse import coo_matrix, csr_matrix, dok_matrix

import scipy.sparse.linalg
from sortedcontainers import SortedDict, SortedList
import pdb

ZIT_EPS = 10**-9
ZIT_NB = 10**5
TEST_CONN = True

########################################################
################# READING
def readParameters(fp=None):
    tf_dict = {"True":True, "False":False, "true":True, "false": False}
    noparse = ["ks", "lqs", "Qs", "ownopiws"]

    ### pickle: -1 disabled, 0 store, 1 load
    ### log: - stdout, + auto filename, * both, other filename
    ### normalize comp: flip comparison edges A>B to A<B
    ### try simply: try finding simpler clauses when remapping 
    ### try V: which shapes to try 1 /\, 2 \/, 3 -- positive indices: even if simple edge has enough support
    org_parameters = {'verbosity': 3, 'logfile': "+", "ego": -1, "center":"",
                      'results_rep': "../results/", "data_rep":"../data/", "labels_ext": "labels",
                      "action": "team", "graph_ext": "graph", "ks":"10", "lqs":"3", "Qs": None, "output": "+",
                      "nba": -10, "nbpoints": 3, "ampl": 0.1, "backtrack": 0, "ownopiws": "0",
                      "min_count": 1.0, "min_weight": None, "opityp": "l", "nbo": 1}

    # 'limit_paths_itemsets': 1000, 'limit_graphs_itemsets': 1000,
    # 'limit_unit_instances': 0, 'limit_abs_instances': 10**5,

    if fp is not None:
        read_parameters = {}
        for line in fp:
            if re.match("^ *#", line) is not None:
                continue
            parts = line.split()
            if len(parts) > 1:
                if parts[0] in noparse:
                    tmp = " ".join(parts[1:])
                    read_parameters[parts[0]] = tmp
                elif parts[1] in tf_dict:
                    read_parameters[parts[0]] = tf_dict[parts[1]]
                else:
                    try:
                        read_parameters[parts[0]] = int(parts[1])
                    except ValueError:
                        try:
                            read_parameters[parts[0]] = float(parts[1])
                        except ValueError:
                            if "~" in parts[1]:
                                tmp = re.sub("~", os.path.expanduser("~"), parts[1])
                            else:
                                tmp = " ".join(parts[1:])
                            read_parameters[parts[0]] = tmp
        org_parameters.update(read_parameters)

    org_parameters["ks"] = [int(s) for s in org_parameters["ks"].strip().split(",")]
    org_parameters["ownopiws"] = [float(s) for s in org_parameters["ownopiws"].strip().split(",")]
    org_parameters["lqs"] = [int(s) for s in org_parameters["lqs"].strip().split(",")]
    if org_parameters.get("Qs", None) is not None: 
        org_parameters["Qs"] = org_parameters["Qs"].strip().split(",")
    return org_parameters

def readGraph(filename, min_weight=None, center=None):
    graph = {}
    # with codecs.open(filename, encoding='utf-8', mode='r') as f:
    sep = None
    counts = 0
    with open(filename, mode='r') as f:
        for line in f:
            if re.match("\s*#", line):
                continue
            # #### WARNING ONLY DEBUG!
            # counts += 1
            # if counts > 1000:
            #     break

            if sep is None:
                tmp = re.match("^(?P<from>[^\t]*)\t(?P<to>[^\t]*)(\t(?P<count>[0-9.]*) *)?$", line.strip())
                if tmp is None:
                    tmp = re.match("^(?P<from>[^,]*),(?P<to>[^,]*)(,(?P<count>[0-9.]*) *)?$", line.strip())
                    if tmp is None:
                        sep = " "
                    else:
                        sep = ","
                else:
                    sep = "\t"
            tmp = re.match("^(?P<from>[^"+sep+"]*)"+sep+"(?P<to>[^"+sep+"]*)("+sep+"(?P<count>[0-9.]*) *)?$", line.strip())
            if tmp is not None and tmp.group("from") != tmp.group("to"):
                nfrom, nto, ncount = tmp.group("from"), tmp.group("to"), None
                if tmp.group("count") is not None:
                    ncount = float(tmp.group("count"))
                if ncount >= min_weight and center not in [nfrom, nto]:
                    for (f,t) in [(nfrom, nto), (nto, nfrom)]:
                        if not graph.has_key(f):
                            graph[f] = {t:ncount}
                        else:
                            graph[f][t] = ncount
    return graph

def largestComponent(graph):
    G = nkx.Graph(itertools.chain(*[[(nA,nB) for nB in neighs.keys()] for nA, neighs in graph.items()]))
    ccc = list(nkx.connected_components(G))
    if len(ccc) > 0:
        ccc.sort(key =lambda x: len(x))
        nodes_keep = set(ccc[-1])
        new_graph = dict([(nA, dict([(nB, graph[nA][nB]) for nB in nodes_keep.intersection(graph[nA].keys())])) for nA in nodes_keep])
        return new_graph
    return graph

def readLocs(lfilename, graph, min_cl=0):
    if os.path.isfile(lfilename):
        if lfilename.split(".")[-1] == "checkins":
            return readLocationsC(lfilename, graph, min_cl), "coo"
        if lfilename.split(".")[-1] in ["csv", "locations"]:
            return readLocationsL(lfilename, graph), "coo"
    return None, "-"

def readLabels(lfilename, graph, min_cl=0):
    if os.path.isfile(lfilename):
        if lfilename.split(".")[-1] == "checkins":
            return readLocationsC(lfilename, graph, min_cl), "coo"
        if lfilename.split(".")[-1] == "locations":
            return readLocationsL(lfilename, graph), "coo"
        if lfilename.split(".")[-1] == "feat":
            return readLabelsF(lfilename, graph, min_cl), "lbl"
        if lfilename.split(".")[-1] == "featlabels":
            return readLabelsL(lfilename, graph, min_cl), "lbl"
        else:
            return readLabelsW(lfilename, graph, min_cl), "lbl"
    return {}, "none"

def readLocationsL(lfilename, graph):
    ndloc_dict = {}
    patt = None
    with open(lfilename, mode='r') as f:
        for line in f:
            if patt is None:
                tmp = re.match("^(?P<node>\S+)\s+(?P<llat>[-+.0-9]+)\s+(?P<llong>[-+.0-9]+)$", line.strip())
                if tmp is None:
                    tmp = re.match("^(?P<node>[^,]+),(?P<llat>[-+.0-9]+),(?P<llong>[-+.0-9]+)$", line.strip())
                    if tmp is not None:
                        patt = "^(?P<node>[^,]+),(?P<llat>[-+.0-9]+),(?P<llong>[-+.0-9]+)$"
                else:
                    patt = "^(?P<node>\S+)\s+(?P<llat>[-+.0-9]+)\s+(?P<llong>[-+.0-9]+)$"
            if patt is not None:
                tmp = re.match(patt, line)
                if tmp is not None and ( graph is None or graph.has_key(tmp.group("node"))) and \
                       (float(tmp.group("llong")), float(tmp.group("llat"))) != (0,0):
                #### So that they are in the right order from plotting Graph
                    # ndloc_dict[tmp.group("node")] = (float(tmp.group("llat")), float(tmp.group("llong")))
                    ndloc_dict[tmp.group("node")] = (float(tmp.group("llong")), float(tmp.group("llat")))
    return ndloc_dict
    
def readLocationsC(lfilename, graph, typ=0):
    ndloc_dict = {}
    counts = 0
    with open(lfilename, mode='r') as f:
        for line in f:
            # #### WARNING ONLY DEBUG!
            # counts += 1
            # if counts > 1000:
            #     break
            
            tmp = re.match("^(?P<node>\S+)\s+(?P<timestamp>\S+)\s+(?P<llat>[-+.0-9]+)\s+(?P<llong>[-+.0-9]+)\s+(?P<locid>[0-9a-zA-Z]+)$", line)
            if tmp is not None and ( graph is None or graph.has_key(tmp.group("node"))):
                nd = tmp.group("node")
                if nd in ndloc_dict:
                    if tmp.group("locid") in ndloc_dict[nd]:
                        ndloc_dict[nd][tmp.group("locid")].append((tmp.group("timestamp"), float(tmp.group("llat")), float(tmp.group("llong"))))
                    else:
                        ndloc_dict[nd][tmp.group("locid")] = [(tmp.group("timestamp"), float(tmp.group("llat")), float(tmp.group("llong")))]
                else:
                    ndloc_dict[nd] = {tmp.group("locid"): [(tmp.group("timestamp"), float(tmp.group("llat")), float(tmp.group("llong")))]}
    return prepareLocations(ndloc_dict, typ)

def prepareLocations(ndloc_dict, typ=0):
    #if typ == 0:
    return prepareLocationsTop(ndloc_dict)

def prepareLocationsTop(ndloc_dict):
    nodes = ndloc_dict.keys()
    for node in nodes:
        k = sorted(ndloc_dict[node].keys(), key = lambda x: len(ndloc_dict[node][x]))[-1]
        ndloc_dict[node] = ndloc_dict[node][k][0][1:3]
    return ndloc_dict


def readLabelsF(lfilename, graph=None, min_cl=0):
    ndlb_dict = {}
    names = []
    with open(lfilename+"names", mode='r') as f:
        for line in f:
            names.append(line.strip())

    with open(lfilename, mode='r') as f:
        for line in f:
            tmp = re.match("^(?P<node>[^ ]*) (?P<feats>[01 \t]*)$", line)
            if tmp is not None and ( graph is None or graph.has_key(tmp.group("node"))):
                # ndlb_dict[tmp.group("node")] = dict([("%d" % i,1) for i,v in enumerate(tmp.group("feats").strip().split()) if v != "0"])
                ndlb_dict[tmp.group("node")] = dict([(names[i],1) for i,v in enumerate(tmp.group("feats").strip().split()) if v != "0"])
    return ndlb_dict

def readLabelsL(lfilename, graph=None, min_cl=0):
    ndlb_dict = {}
    with open(lfilename, mode='r') as f:
        for line in f:
            tmp = re.match("^(?P<node>[^ ]*) (?P<term>.*)$", line)
            if tmp is not None and ( graph is None or graph.has_key(tmp.group("node"))):
                if ndlb_dict.has_key(tmp.group("node")):
                    ndlb_dict[tmp.group("node")][tmp.group("term").strip()] = 1
                else:
                    ndlb_dict[tmp.group("node")] = {tmp.group("term").strip(): 1}
    return ndlb_dict

def readLabelsW(lfilename, graph=None, min_cl=0):
    ndlb_dict = {}
    cnode = None
    with open(lfilename, mode='r') as f:
        for line in f:
            tmp = re.match("^(?P<node>[^\t]*)\t(?P<term>[^\t]*)\t(?P<count>\d*)$", line)
            if tmp is not None and ( int(tmp.group("count")) >= min_cl ) and ( graph is None or graph.has_key(tmp.group("node"))):
                if ndlb_dict.has_key(tmp.group("node")):
                    ndlb_dict[tmp.group("node")][tmp.group("term")] = int(tmp.group("count"))
                else:
                    ndlb_dict[tmp.group("node")] = {tmp.group("term"): int(tmp.group("count"))}
    return ndlb_dict

def readLabelsS(lfilename, graph=None, min_cl=0):
    stop_words = "algorithm analysy applicat approach base comput control data design gener informat method model multy network study system time".split()
    ndlb_dict = {}
    cnode = None
    with open(lfilename, mode='r') as f:
        for line in f:
            tmp = re.match("^(?P<node>[^\t]*)\t(?P<term>[^\t]*)\t(?P<count>\d*)$", line)
            if tmp is not None and ( int(tmp.group("count")) >= min_cl ) and ( graph is None or graph.has_key(tmp.group("node"))):
                if tmp.group("node") != cnode and cnode is not None:
                    sm = float(sum(ndlb_dict[cnode].values()))
                    keep = set(ndlb_dict[cnode].keys()) - remove
                    ndlb_dict[cnode] = dict([(k, ndlb_dict[cnode][k]) for k in keep if ndlb_dict[cnode][k]/sm >= data_parameters["min_frac"] and ndlb_dict[cnode][k] >= data_parameters["min_count"]])
                    cnode = tmp.group("node")
                    
                if tmp.group("term") not in stop_words:
                    if ndlb_dict.has_key(tmp.group("node")):
                        ndlb_dict[tmp.group("node")][tmp.group("term")] = int(tmp.group("count"))
                    else:
                        ndlb_dict[tmp.group("node")] = {tmp.group("term"): int(tmp.group("count"))}
    return ndlb_dict

def readClusts(lfilename, graph):
    clusts = []
    with open(lfilename, "r") as fp:
        for line in fp:
            clusts.append(set())
            parts = line.strip().split("\t")
            if len(parts) > 1:
                for p in parts[1].split(")"):
                    if len(p) > 2:
                        node  = p.split("#")[0].strip()
                        if node in graph:
                            clusts[-1].add(node)
    return clusts

def readClustsSimple(lfilename, graph):
    clusts = []
    with open(lfilename, "r") as fp:
        for line in fp:
            clusts.append(set(line.strip().split(", ")))
    return clusts

            
def getNodesToLabels(graph, ndlb_dict, data_parameters):
    nodes_to_labels = {}
    for node, lbd in ndlb_dict.items():
        nodes_to_labels[node] = set(lbd.keys())
    return nodes_to_labels

def readCandidates(lfilename, lsorted):
    candidates = []
    with open(lfilename, mode='r') as f:
        for line in f:
            candidates.append([lsorted[int(p)] for p in line.strip().split()[:-1]])
    return candidates

def readClusters(lfilename, uls):
    clusters = []
    with open(lfilename, mode='r') as f:
        for line in f:
            try:
                clusters.append(set([uls[int(pi)-1] for pi in line.split()]))
            except IndexError:
                clusters.append(set())
    return clusters

def readCLabels(lfilename):
    labelsets = []
    with open(lfilename, mode='r') as f:
        for line in f:
            labelsets.append(sorted([l.strip() for l in line.split()]))
    return labelsets

##### make opinions correlated with locations
def flocops(locs, nbo, data_parameters):
    N = locs.shape[1]
    nopis = []
    aopis = []
    ans = []
    minc, maxc = (numpy.min(locs, axis=1), numpy.max(locs, axis=1))
    corners = numpy.vstack([[maxc[0], maxc[1]], [minc[0], minc[1]], [minc[0], maxc[1]], [maxc[0], minc[1]]]).T
    if data_parameters["nba"] < 0:
        nba = N/abs(data_parameters["nba"])
    else:
        nba = data_parameters["nba"]
    nbpoints = data_parameters["nbpoints"] # 0
    ampl = data_parameters["ampl"]
    for n in range(nbo):
        ans.append(numpy.random.choice(locs.shape[1], nba))
        anchors = numpy.hstack([locs[:, ans[-1]], corners])
        ccc = scipy.spatial.distance.cdist(locs.T, anchors.T, metric="euclidean")
        zzz = numpy.where(ccc==0)
        ccc[zzz] = 1
        csim = 1/ccc
        vv = []
        comp_dists = []
        for vk in range(nbpoints):
            vv.append(numpy.argmax(csim, axis=1))
            comp_dists.append(csim[range(N), vv[-1]])
            csim[range(N), vv[-1]]  = 0
        ttt_dists = numpy.vstack(comp_dists)
        ttt_dists /= (numpy.outer(numpy.ones(nbpoints), numpy.sum(ttt_dists, axis=0)))
        ttt = numpy.vstack(vv)
        aopis.append(numpy.round(numpy.random.rand(nba+4)))
        # aopis.append(numpy.random.rand(nba+4))
        tmpo = numpy.sum(aopis[-1][ttt]*ttt_dists, axis=0)
        if ampl > 0:
            tmpo += ampl*(numpy.random.rand(N)-0.5)
            tmpo += ampl/2.0
            tmpo /= 1+ampl
        # tmpo[numpy.where(tmpo<0)] = 0
        # tmpo[numpy.where(tmpo>1)] = 1
        tmpo[ans[-1]] = aopis[-1][:-4]
        nopis.append(tmpo)
    nopis = numpy.vstack(nopis)
    return nopis, aopis, ans

def flocop(loc, opi_centers):
    scipy.spatial.distance.cdist(tmp.T, tmp[:,pivots].T, metric=typ["metric"])
    return ((loc[0]-boundaries[0][0])/(boundaries[0][1]-boundaries[0][0])) + \
           ((loc[1]-boundaries[1][0])/(boundaries[1][1]-boundaries[1][0]))


def geoToOpi(lfilename, graph, nbo, data_parameters):
    locs = {}
    if os.path.isfile(lfilename):
        if lfilename.split(".")[-1] == "checkins":
            locs = readLocationsC(lfilename, graph)
        if lfilename.split(".")[-1] in ["csv", "locations"]:
            locs = readLocationsL(lfilename, graph)
    ks = locs.keys()
    dict_nds = dict([(nd, ii) for (ii, nd) in enumerate(ks)])
    nlocs = numpy.array([locs[n] for n in ks]).T
    opis, aopis, ans = flocops(nlocs, nbo, data_parameters)
    ansk = [[ks[ii] for ii in ani] for ani in ans]
    # opis = numpy.array([[flocop(locs[n], boundaries) for jj in range(nbo)] for n in ks]).T
    return opis, dict_nds, locs, aopis, ansk

def lblsToOpi(lfilename, nbo, counts=True):
    if os.path.isfile(lfilename):
        if lfilename.split(".")[-1] in ["feat"]:
            return lblsToOpiF(lfilename, nbo, counts)
        else:
            return lblsToOpiW(lfilename, nbo, counts)
    return None, {}


def lblsToOpiF(lfilename, nbo, counts=False):
    pattern = "^((?P<node>[^ ]*) )?(?P<feats>[0-9 \t]*)$"
    data_rows = []
    dict_auts = {}
    with open(lfilename, mode='r') as f:
        li = -1
        for line in f:
            li += 1
            tmp = re.match(pattern, line)
            if tmp is not None:
                if tmp.group("node") is not None:
                    dict_auts[tmp.group("node")] = li 
                else:
                    dict_auts["%s" % li] = li
                if counts:
                    data_rows.append(map(int, tmp.group("feats").split()))
                else:
                    data_rows.append([1*(int(v)!=0) for v in tmp.group("feats").split()])
    mat_cooc = numpy.array(data_rows)
    tmp = decomposition.TruncatedSVD(n_components=nbo).fit_transform(mat_cooc)
    parts = numpy.vstack([(tmp[:,i]-numpy.min(tmp[:,i]))/(numpy.max(tmp[:,i])-numpy.min(tmp[:,i])) for i in range(tmp.shape[1])])
    return parts, dict_auts

def lblsToOpiW(lfilename, nbo, counts=True): #False):
    patterns= [ "^(?P<node>[^\t]*)\t(?P<term>[^\t]*)(\t(?P<count>\d*))?$",
                "^(?P<node>[^ ]*) (?P<term>.*)( (?P<count>\d*))?$"]
    if lfilename.split(".")[-1] == "nfeatlabels":
        patterns= [ "^(?P<node>[^ ]*) (?P<term>[^ ]*)( .*)?((?P<count>\d*))?$"]

    patper = None
    tmp = None
    entries_list = [[],[],[]]
    with open(lfilename, mode='r') as f:
        for line in f:
            #### DETERMINE WHICH PATTERN TO USE
            if patper is None:
                patper = 0
                while tmp is None and patper < len(patterns):
                    tmp = re.match(patterns[patper], line)
                    patper += 1
                if tmp is None:
                    return None, {}
                else:
                    patper -= 1
                
            tmp = re.match(patterns[patper], line)
            if tmp is not None:
                entries_list[0].append(tmp.group("node"))
                entries_list[1].append(tmp.group("term").strip())
                # if counts and tmp.group("count") is not None:
                #     entries_list[2].append(int(tmp.group("count")))

    dict_auts = dict([(a,i) for i,a in enumerate(set(entries_list[0]))])
    dict_wds = dict([(w,i) for i,w in enumerate(set(entries_list[1]))])
    if counts and len(entries_list[2]) == len(entries_list[1]):
        mat_cooc = csr_matrix( (entries_list[2],([dict_auts[i] for i in entries_list[0]], [dict_wds[j] for j in entries_list[1]])))
    else:
        mat_cooc = csr_matrix( (numpy.ones(len(entries_list[0])),([dict_auts[i] for i in entries_list[0]], [dict_wds[j] for j in entries_list[1]])))
    tmp = decomposition.TruncatedSVD(n_components=nbo).fit_transform(mat_cooc)
    parts = numpy.vstack([(tmp[:,i]-numpy.min(tmp[:,i]))/(numpy.max(tmp[:,i])-numpy.min(tmp[:,i])) for i in range(tmp.shape[1])])
    return parts, dict_auts

def printOpis(lfilename, parts, dict_auts):
    with open(lfilename, mode='w') as f:
        for d, di in dict_auts.items():
            try:
                f.write("\t".join([d]+["%s" % pi for pi in parts[:,di]])+"\n")
            except OverflowError:
                pdb.set_trace()
                print d
            
def readOpis(lfilename):
    dict_auts = {}
    parts_lst = []
    with open(lfilename, mode='r') as f:
        for line in f:
            tmp = line.strip().split("\t")
            dict_auts[tmp[0]] = len(parts_lst)
            parts_lst.append(map(float, tmp[1:]))
    return numpy.array(parts_lst).T, dict_auts
            
def cleanCache(data_parameters):
    for v in ["store_utils"]:
        if data_parameters.has_key(v):
            del data_parameters[v]

def cleanUpGraph(graph, nodes):
    for node in nodes:
        for neigh in graph[node].keys():
            del graph[neigh][node]
        del graph[node]


def readGraphNum(filename, graph=None, filter_set=None):
    if graph is None:
        graph = {}
    with open(filename, mode='r') as f:
        for line in f:
            if re.match("\s*#", line):
                continue

            ppt = map(int, line.strip().split())
            if filter_set is None or ppt[0] in filter_set or ppt[1] in filter_set:
                for (f,t) in [(0,1), (1,0)]:
                    if not graph.has_key(ppt[f]):
                        graph[ppt[f]] = {ppt[t]: 1}
                    else:
                        graph[ppt[f]][ppt[t]] = 1
    return graph

def readOpisNum(lfilename):
    parts_lst = []
    with open(lfilename, mode='r') as f:
        for line in f:
            parts_lst.append(map(float, line.strip().split()))
    return numpy.array(parts_lst)

def readQsNum(lfilename):
    Qq = []
    Qdists = []
    with open(lfilename, mode='r') as f:
        for line in f:
            parts = line.split("\t")
            Qq.append(map(int, parts[0].split()))
            Qdists.append((map(int, parts[1].split()), map(int, parts[2].split())))
    return Qq, Qdists

########################################################
################# PRINTING
def openMakeP(filename, mode="w"):
    d = os.path.dirname(filename)
    if not os.path.isdir(d):
        os.makedirs(d)
    return open(filename, mode)

def initResOut(data_parameters, external, k):
    if data_parameters["output"] != "-":
        if data_parameters["output"] == "+":
            fbasis = "%s%s.%s_%s" % (data_parameters['results_rep'], data_parameters['basis'], external, k)
            fo = openMakeP(fbasis+".communities", "w")
            fe = openMakeP(fbasis+".cclust", "w")
        else:
            fbasis = data_parameters["output"]
            fo = openMakeP(fbasis+".communities", "w")
            fe = fo
    else:
        fo = sys.stdout
        fe = None
    fo.write("------------METHOD=%s k=%s\n" % (external, k))
    return fo, fe

class Dummyfp:
    def write(self, txt):
        pass
    def close(self):
        pass
    
def initResOutSimple(data_parameters, external, k):
    fe = None
    if data_parameters.get("output", "-") == "-":
        return Dummyfp(), Dummyfp()
    if data_parameters.get("output", "-") != "-":
        if data_parameters["output"] == "+":
            fbasis = "%s%s.%s_%s" % (data_parameters['results_rep'], data_parameters['basis'], external, k)
            fo = openMakeP(fbasis+".communities", "w")
            # fe = openMakeP(fbasis+".cclust", "w")
        else:
            fbasis = data_parameters["output"]
            fo = openMakeP(fbasis+".communities", "w")
            # fe = fo

    else:
        fo = sys.stdout
        fe = None
    fo.write("------------METHOD=%s k=%s\n" % (external, k))
    return fo, fe

def closeOut(fo):
    if fo is not None and fo != sys.stdout:
        fo.close()


def labelsOut(nodes_to_labels, lmap, fname_mtx):
    fo = openMakeP(fname_mtx, "w")
    for e in sorted(nodes_to_labels.keys()):
        fo.write(" ".join([lmap[l] for l in nodes_to_labels[e]])+"\n")
    fo.close()

def edgesOut(graph, fname_grl, fname_grd):
    uls = sorted([k for k,vs in graph.items() if len(vs) > 0])
    fo = openMakeP(fname_grl, "w")
    fo.write("\n".join(uls))
    fo.close()

    lsdict = dict([(v,i+1) for i,v in enumerate(uls)])
    
    fo = openMakeP(fname_grd, "w")
    for k, vs in graph.items():
        fo.write(("\n".join(["%d %d" % (ldict[k], ldict[t]) for t in vs.keys()]))+"\n")
    fo.close()
    return uls

def graphOut(graph, fname_grl, fname_grd, weighted=False):
    uls = sorted([k for k,vs in graph.items() if len(vs) > 0])
    fo = openMakeP(fname_grl, "w")
    fo.write("\n".join(uls))
    fo.close()

    lsdict = dict([(v,i+1) for i,v in enumerate(uls)])
    
    fo = openMakeP(fname_grd, "w")
    if weighted:
        fo.write("%d %d 001\n" % (len(lsdict), sum([len(k) for k in graph.values()])/2))
    else:
        fo.write("%d %d\n" % (len(lsdict), sum([len(k) for k in graph.values()])/2))
    for k, vs in graph.items():
        if weighted:
            tmp = sorted([(lsdict[v], 1000*w) for v, w in vs.items()])
            fo.write((" ".join(["%d %d" % t for t in tmp]))+"\n")
        else:
            tmp = sorted([lsdict[v] for v, w in vs.items()])
            fo.write((" ".join(["%d" % t for t in tmp]))+"\n")
    fo.close()
    return uls

def externalOut(graph, nodes_to_labels, basis):    
    nls = getNodeOrds(graph)

    #### GRAPH EDGES LIST AND METIS FORMAT
    lsdict = dict([(v,i+1) for i,v in enumerate(nls)])    
    fe = openMakeP(basis+".ledges", "w")
    fo = openMakeP(basis+".metis", "w")
    fo.write("%d %d\n" % (len(lsdict), sum([len(k) for k in graph.values()])/2))
    for k, vs in graph.items():
        tmp = sorted([lsdict[v] for v, w in vs.items()])
        fo.write((" ".join(["%d" % t for t in tmp]))+"\n")
        for t in tmp:
            if t > lsdict[k]:
                fe.write("%d %d\n" % (lsdict[k], t))
    fo.close()
    fe.close()

    ### LABEL MATRIX
    lbls = set()
    for vs in nodes_to_labels.values():
        lbls.update(vs)
    lbls = sorted(lbls)

    ### LABEL MAP
    fo = openMakeP(basis+".lmap", "w")
    fo.write("\n".join(lbls))
    fo.close()

    lbdict = dict([(v,"%d" % (i+1)) for i,v in enumerate(lbls)])
    fo = openMakeP(basis+".lmatrix", "w")
    fo.write("%d\n" % len(nls))
    fo.write("%d\n" % len(lbls))
    for u in nls:
        fo.write("%d " % len(nodes_to_labels.get(u, [])))
        fo.write((" ".join(sorted([lbdict[l] for l in nodes_to_labels.get(u, [])])))+"\n")
    fo.close()
    return nls

def DCMOut(graph, nodes_to_labels, basis):    
    nls = getNodeOrds(graph)
    lsdict = dict([(v,i) for i,v in enumerate(nls)])    
    nkeys = sorted([(lsdict[k], k) for k in graph.keys()])

    ### LABEL MATRIX
    lbls = set()
    for vs in nodes_to_labels.values():
        lbls.update(vs)
    lbls = sorted(lbls)
    lbdict = dict([(v,i) for i,v in enumerate(lbls)])

    ### WRITE OUT NODE MAP
    fo = openMakeP(basis+".nodes_map", "w")
    fo.write("\n".join([n[1] for n in nkeys]))
    fo.close()

    ### WRITE OUT LABEL MAP
    fo = openMakeP(basis+".labels_map", "w")
    fo.write("\n".join(lbls))
    fo.close()

   #### GRAPH EDGES LIST AND METIS FORMAT
    fe = openMakeP(basis+".dcm_edges", "w")
    fe.write("%d\n" % len(lsdict))
    for k,n in nkeys:
        tmp = sorted([lsdict[v] for v, w in graph[n].items()])
        fe.write((" ".join(["%d" % t for t in [k]+tmp]))+"\n")
    fe.close()

    fe = openMakeP(basis+".blda_edges", "w")
    for k,n in nkeys:
        tmp = sorted([lsdict[v] for v, w in graph[n].items() if lsdict[v] < k])
        for ko in tmp:
            fe.write("2 %d %d\n" % (k, ko))
    fe.close()


    fo = openMakeP(basis+".dcm_labels", "w")
    for u in nls:
        tmp = sorted([lbdict[l] for l in nodes_to_labels.get(u, [])])
        fo.write((" ".join(["%d" % t for t in tmp]))+"\n")
    fo.close()

    labels_to_nodes = reverseNdlbSets(nodes_to_labels)
    fo = openMakeP(basis+".blda_labels", "w")
    for l in lbls:
        tmp = sorted([lsdict[n] for n in labels_to_nodes.get(l, [])])
        fo.write(("%d " % len(tmp)) + (" ".join(["%d" % t for t in tmp]))+"\n")
    fo.close()
    return nls


def addGroup(ti, scr, lspec, edgeset, nodeset, labelset=None, covered=None, coveredn=None, nb_e=-1, nb_n=-1, all_edges=set(), graph=None):
    if covered is not None:
        if type(covered) == dict:
            newedges = edgeset.difference(covered.keys())
        else:
            newedges = edgeset.difference(covered)
    if coveredn is not None:
        if type(covered) == dict:
            newnodes = nodeset.difference(coveredn.keys())
        else:
            newnodes = nodeset.difference(coveredn)
    nn = len(nodeset)
    
    if type(covered) == dict:
        for edge in newedges:
            covered[edge] = ti
        for node in nodeset:
            if coveredn.has_key(node):
                coveredn[node].append(ti)
            else:
                coveredn[node]= [ti]
    elif covered is not None:
        covered |= newedges
        coveredn |= newnodes

    tmp = ("*(%d)\ts=%f sL=%f d=%d dR=%f E=%d N=%d L=%d" % (ti, scr, lspec, density(edgeset, nn), densityR(edgeset, nn), len(edgeset), nn, len(labelset)))
    if covered is not None and coveredn is not None:
        tmp += (" Cd=%d CdR=%f CE=%d CN=%d" % (density(newedges, nn), densityR(newedges, nn), len(newedges), len(newnodes)))
        tmp += (" (%d/%d=%f, %d/%d=%f)"  % (len(covered), nb_e, len(covered)/nb_e, len(coveredn), nb_n, len(coveredn)/nb_n))
        tmp += (" cut_edges=%d overlap_nodes=%f"  % (len([e for e in all_edges - set(covered.keys()) if coveredn.has_key(e[0]) and coveredn.has_key(e[1])]), numpy.mean([len(v) for v in coveredn.values()])))
    if labelset is not None:
        tmp += "\n[%s]" % ", ".join(labelset)

    # if graph is not None:
    #     # tmp += ("\n (%s)\n" % (", ".join(["%s [%.2f]" % nif for nif in sorted([(n, sum(graph[n].values())) for n in nodeset], key=lambda x: x[1], reverse=True)])))
    #     tmp += ("\n (%s)\n" % (", ".join(sorted(nodeset))))
    if graph is not None and len(nodeset) > 5:
        # tmp += ("\n (%s ...)\n" % (", ".join(sorted(random.sample(nodeset,5)))))
        tmp += ("\n (%s ...)\n" % (", ".join(sorted(nodeset)[:6])))
    else:
        tmp += ("\n (%s)\n" % (", ".join(sorted(nodeset))))
    return tmp, newedges, newnodes

########################################################
################# MISC
def ratioZ(a,b):
    if b == 0:
        if a == 0:
            return 0.0
        return float("Inf")
    else:
        return (b-a)/b

def divZ(a,b):
    if b == 0:
        return float("Inf")
    else:
        return a/b
    
def reverseNdlb(ndlb_dict, nodes=None):
    labels_to_nodes = {}
    if nodes is None:
        nodes = ndlb_dict.keys()
    for n in nodes:
        for vi in ndlb_dict.get(n,{}).keys():
            if labels_to_nodes.has_key(vi):
                labels_to_nodes[vi].add(n)
            else:
                labels_to_nodes[vi] = set([n])
    return labels_to_nodes

def reverseNdlbSets(ndlb_dict, nodes=None):
    labels_to_nodes = {}
    if nodes is None:
        nodes = ndlb_dict.keys()
    for n in nodes:
        for vi in ndlb_dict.get(n,set()):
            if labels_to_nodes.has_key(vi):
                labels_to_nodes[vi].add(n)
            else:
                labels_to_nodes[vi] = set([n])
    return labels_to_nodes


def reverseLabels(nodes_to_labels, nodes=None):
    labels_to_nodes = {}
    if nodes is None:
        nodes = nodes_to_labels.keys()
    for n in nodes:
        for vi in nodes_to_labels.get(n,[]):
            if labels_to_nodes.has_key(vi):
                labels_to_nodes[vi].add(n)
            else:
                labels_to_nodes[vi] = set([n])
    return labels_to_nodes

def getReversedLabels(nodes_to_labels, graph, data_parameters, store=True):
    if not data_parameters.has_key("labels_to_nodes") or not store:
        labels_to_nodes = reverseLabels(nodes_to_labels, graph.keys())
        if store:
            data_parameters["labels_to_nodes"] = labels_to_nodes
    else:
        labels_to_nodes = data_parameters["labels_to_nodes"]
    return labels_to_nodes

def getNodeset(edgeset):
    return set(itertools.chain.from_iterable(zip(*edgeset)))

def getNodeOrds(graph):
    return sorted([k for k,vs in graph.items() if len(vs) > 0]) + sorted([k for k,vs in graph.items() if len(vs) == 0])

# def getNodeId(node, graph, data_parameters):
#     if not data_parameters.has_key("lsdict"):
#         lsdict = dict([(v,"%d" % (i+1)) for i,v in enumerate(getNodeOrds(graph))])
#         data_parameters["lsdict"] = lsdict
#     else:
#         lsdict = data_parameters["lsdict"]
#     return lsdict[node]

### to print the node names rather than num id
def getNodeId(node, graph, data_parameters):
    return node
        
def projectGraph(nodes, graph, min_weight=None):
    subgraph = {}
    for node in nodes:
        subgraph[node] = dict([(nt, nc) for nt, nc in graph.get(node, {}).items() if nc >= min_weight])
        if len(subgraph[node]) == 0:
            del subgraph[node]
    return subgraph

def makeEdgesCands(graph, nodes_to_labels):
    labels_to_edges = {}
    edges_to_labels = {}
    for nf, nts in graph.items():
        for nt, c in nts.items():
            if nf == min(nf, nt):
                tlbs = nodes_to_labels.get(nf, set()) & nodes_to_labels.get(nt, set())
                edges_to_labels[(nf,nt)] = tlbs
                for l in tlbs:
                    if labels_to_edges.has_key(l):
                        labels_to_edges[l].add((nf,nt))
                    else:
                        labels_to_edges[l] = set([(nf,nt)])
    return labels_to_edges, edges_to_labels

def getEdgesCands(nodes_to_labels, graph, data_parameters, store=True):
    if not data_parameters.has_key("labels_to_edges") or not store:
        labels_to_edges, edges_to_labels = makeEdgesCands(graph, nodes_to_labels)
        if store:
            data_parameters["labels_to_edges"], data_parameters["edges_to_labels"] = labels_to_edges, edges_to_labels
    else:
        labels_to_edges, edges_to_labels = data_parameters["labels_to_edges"], data_parameters["edges_to_labels"]
    return labels_to_edges, edges_to_labels
        
def cleanLabels(ndlb_dict, data_parameters):
    for a in ndlb_dict.keys():
        sm = float(sum(ndlb_dict[a].values()))
        ndlb_dict[a] = dict([k for k in ndlb_dict[a].items() if k[1]/sm >= data_parameters["min_frac"] and k[1] >= data_parameters["min_count"]])

def normalizeLabelsCounts(ndlb_dict, data_parameters):
    remove = set()
    if data_parameters.has_key("stop_frac"):
        nbstop = len(ndlb_dict)*data_parameters["stop_frac"]
        labels_to_nodes = reverseNdlb(ndlb_dict)
        remove = set([l for l,v in labels_to_nodes.items() if len(v) >= nbstop])

    for a in ndlb_dict.keys():
        sm = float(sum(ndlb_dict[a].values()))
        keep = set(ndlb_dict[a].keys()) - remove
        ndlb_dict[a] = dict([(k, ndlb_dict[a][k]) for k in keep if ndlb_dict[a][k]/sm >= data_parameters["min_frac"] and ndlb_dict[a][k] >= data_parameters["min_count"]])


def updateSpreads(graph, ndlb_dict, labelset, cnodes, labels_to_nodes, labels_to_edges, labels_to_nodes_unspread, data_parameters):
#    print labelset
#    pdb.set_trace()
    for label in labelset:
        up_nodes = set([node for node in labels_to_nodes[label] if (node not in cnodes) and len(labels_to_nodes_unspread[label].intersection(graph.get(node, {}).keys()).difference(cnodes)) > len(graph.get(node, {}).keys())/data_parameters["spread_thres"]])
        up_edges = set([edge for edge in labels_to_edges[label] if (edge[0] in up_nodes) and (edge[1] in up_nodes)])
        labels_to_nodes[label] = up_nodes
        labels_to_edges[label] = up_edges

def spreadLabels(graph, ndlb_dict, data_parameters):
    all_nodes = set(graph.keys()).union(ndlb_dict.keys())
    spread_ndlb_dict = {}
    for node in all_nodes:
        neighs = graph.get(node, {})
        spread_ndlb_dict[node] = dict(ndlb_dict.get(node, {}))
        cand_labels = {}
        for neigh, nw in neighs.items():
            for label, lw in ndlb_dict.get(neigh, {}).items():
                if data_parameters["spread_weighted"] == 0:
                    lw = 1.0
                if data_parameters["spread_outdiv"] == 1:
                    lw = lw / float(len(graph.get(neigh, {})))
                if data_parameters["spread_outweight"] == 1:
                    lw = lw / float(sum(ndlb_dict[neigh].values()))
                ### options for lw: 1, 1.0/len(graph.get(node, {}))
                if cand_labels.has_key(label):
                    cand_labels[label] += lw
                else:
                    cand_labels[label] = lw

        for label, lw in cand_labels.items():
            if data_parameters["spread_indiv"] == 1:
                lw = lw / float(len(graph.get(node, {})))

        # lbls = set(cand_labels.keys()).difference(spread_ndlb_dict[node])
        # for label in lbls:
        #     if re.search("c:[0-9]", node) and re.search("C[0-9]", label):
        #         print node, label
        #         pdb.set_trace()
        
            if lw >= data_parameters["spread_thres"]:
                if label not in spread_ndlb_dict[node]:
                    spread_ndlb_dict[node][label] = 1
    return spread_ndlb_dict

def extractSubs(graph, nodes_to_labels, data_parameters):
    min_weight = data_parameters["min_weight"]
    min_cl = 0
    nodes = set()
    if data_parameters["ego"] > 0:
        ego = data_parameters["ego"]
    # for ego in [2]:
        for centert in data_parameters["centers"].split(";"):
            center = centert.strip()
            
            if ego > 0 and center is not None and graph.has_key(center):
                print "Filtering %s %d" % (center, ego)
                ### GATHER THE EGO NET
                last_nodes = set([center])
                new_nodes = set()
                for i in range(ego+1):
                    for node in last_nodes:
                        if graph.has_key(node):
                            new_nodes.update([t for t,c in graph[node].items() if c >= min_weight])
                    nodes |= last_nodes
                    last_nodes = new_nodes - nodes
                    new_nodes = set()


            if min_weight is None:
                min_weight_s = ""
            else:
                min_weight_s = "-%.1f" % min_weight
            if "ego" in data_parameters["action"]:
                fo = openMakeP(data_parameters['data_rep'] + ("wegonet_%s%d%s.graph" % (center.replace(" ",""), ego, min_weight_s)), "w")
                for node in nodes:
                    for nt, nc in graph.get(node, {}).items():
                        if nt in nodes and nc >= min_weight:
                            fo.write("%s\t%s\t%f\n" % (node, nt, nc))
                fo.close()


                if data_parameters.has_key('labels_file'):
                    fo = openMakeP(data_parameters['data_rep'] + ("wegonet_%s%d%s.labels" % (center.replace(" ",""), ego, min_weight_s)), "w")
                    with open(data_parameters['data_rep'] + data_parameters['labels_file'], mode='r') as f:
                        for line in f:
                            tmp = re.match("^(?P<node>[^\t]*)\t(?P<term>[^\t]*)\t(?P<count>\d*)$", line)
                            if tmp is not None and ( int(tmp.group("count")) >= min_cl ) and ( nodes is None or tmp.group("node") in nodes):
                                fo.write(line)
                    fo.close()
                nodes = set()

def extractSubsLocs(graph, ndloc_dict, data_parameters):
    min_weight = data_parameters["min_weight"]
    min_cl = 0
    nodes = set()
    if data_parameters["ego"] > 0:
        ego = data_parameters["ego"]
    # for ego in [2]:
        for centert in data_parameters["centers"].split(";"):
            center = centert.strip()
            
            if ego > 0 and center is not None and center in graph and center in ndloc_dict:
                print "Filtering %s %d" % (center, ego)
                ### GATHER THE EGO NET
                last_nodes = set([center])
                new_nodes = set()
                for i in range(ego+1):
                    for node in last_nodes:
                        if graph.has_key(node):
                            new_nodes.update([t for t,c in graph[node].items() if c >= min_weight])
                    nodes |= last_nodes
                    last_nodes = new_nodes - nodes
                    last_nodes.intersection_update(ndloc_dict.keys())
                    new_nodes = set()

                if min_weight is None:
                    min_weight_s = ""
                else:
                    min_weight_s = "-%.1f" % min_weight

                fo = openMakeP(data_parameters['data_rep'] + ("wegonet_%s%d%s.edges" % (center.replace(" ",""), ego, min_weight_s)), "w")
                fol = openMakeP(data_parameters['data_rep'] + ("wegonet_%s%d%s.locations" % (center.replace(" ",""), ego, min_weight_s)), "w")
                for node in nodes:
                    fol.write("%s\t%f\t%f\n" % (node, ndloc_dict[node][0], ndloc_dict[node][1]))
                    for nt, nc in graph.get(node, {}).items():
                        if nt in nodes:
                            fo.write("%s\t%s\n" % (node, nt))
                fo.close()
                fol.close()
                nodes = set()

def filterGLocs(graph, ndloc_dict, data_parameters):
    mind, maxd = (2, 200)
    
    dropn = set(graph.keys()).difference(ndloc_dict.keys())
    for dn in dropn:
        for neigh in graph[dn].keys():
            del graph[neigh][dn]
        del graph[dn]

    print "NOTH done ", len(graph)
    dropn = [n for n in graph.keys() if len(graph[n]) > maxd]
    for dn in dropn:
        for neigh in graph[dn].keys():
            del graph[neigh][dn]
        del graph[dn]
        del ndloc_dict[dn]

    print "MAXD done ", len(graph)
    dropn = [n for n in graph.keys() if len(graph[n]) < mind]
    for dn in dropn:
        for neigh in graph[dn]:
            del graph[neigh][dn]
        del graph[dn]
        del ndloc_dict[dn]
    print "MIND done ", len(graph)
    
    fo = openMakeP(data_parameters['data_rep'] + ("filtered_%d-%d.edges" % (mind, maxd)), "w")
    fol = openMakeP(data_parameters['data_rep'] + ("filtered_%d-%d.locations" % (mind, maxd)), "w")
    for node in graph.keys():
        fol.write("%s\t%f\t%f\n" % (node, ndloc_dict[node][0], ndloc_dict[node][1]))
        for nt, nc in graph.get(node, {}).items():
            fo.write("%s\t%s\n" % (node, nt))
    fo.close()
    fol.close()


def filterGBBox(graph, ndloc_dict, data_parameters):
    bboxes = {"USA" : ((48.504703, -128.360312),
                       (23.628067, -61.915000)),
              "EU": ((66.365554, -27.989219),
                     (36.300649, 35.116250)),
              "East-Asia" : ((-8.772548, 91.542031),
                             (50.725084, 154.823281)),
              "Au-NZ": ((-50.412964, 180),
                        (-9.466782, 91.542031)),
              "Middle-East": ((42.523863, 24.433596),
                              (10.102579, 68.466398)),
              "Latin-America":
              ((27.180287, -119.003904),
               (-55.502634, -34.632139)),
              "Sub-Sahara":
              ((19.107400, -21.709384),
               (-36.770911, 57.743742)),
              "India":
              ((3.297826, 93.545822),
               (32.734995, 60.411056))
              }

    #### COORDS GIVEN arbitrary (LAT, LONG) corners turn to (min, max) in  (LAT, LONG)
    for b in bboxes.keys():
        bboxes[b] = ((min(bboxes[b][0][0], bboxes[b][1][0]),
                      min(bboxes[b][0][1], bboxes[b][1][1])),
                     (max(bboxes[b][0][0], bboxes[b][1][0]),
                      max(bboxes[b][0][1], bboxes[b][1][1])))

    nodesets = dict([(bb, set()) for bb in bboxes.keys()])
    for b, loc in ndloc_dict.items():
        if loc != (0,0):
            for bbox, corners in bboxes.items():
                if loc[0] > corners[0][0] and loc[0] < corners[1][0] and \
                       loc[1] > corners[0][1] and loc[1] < corners[1][1]:
                    nodesets[bbox].add(b)

    for bbox, nds in nodesets.items():
        fo = openMakeP(data_parameters['data_rep'] + ("dbbox_%s.edges" % bbox), "w")
        fol = openMakeP(data_parameters['data_rep'] + ("dbbox_%s.locations" % bbox), "w")
        for node in nds:
            fol.write("%s\t%f\t%f\n" % (node, ndloc_dict[node][0], ndloc_dict[node][1]))
            for nt in nds.intersection(graph.get(node, {}).keys()):
                fo.write("%s\t%s\n" % (node, nt))
        fo.close()
        fol.close()



def parseClusters(graph, nodes_to_labels, clusters, data_parameters, labelsets=None, fo=None, fc=None):
    if fo is None:
        fo = sys.stdout
    labels_to_nodes = getReversedLabels(nodes_to_labels, graph, data_parameters)
    labels_to_edges, edges_to_labels = getEdgesCands(nodes_to_labels, graph, data_parameters)
    all_edges = set(edges_to_labels.keys())

    cids = sorted(range(len(clusters)), key=lambda x: len(clusters[x]))
    covered = {}
    coveredn = {}
    nb_n = float(len(graph))
    nb_e = float(len(edges_to_labels))
    for ti, cid in enumerate(cids):
        nodeset = clusters[cid]
        edgeset = set()
        for node in nodeset:
            try:
                edgeset.update([tuple(sorted([node, nei])) for nei in nodeset.intersection(graph[node].keys())])
            except KeyError:
                pdb.set_trace()
                print node in graph
        if labelsets is None:
            labelset, lspec = chooseLabels(labels_to_nodes, nodeset, data_parameters)
        else:
            labelset = labelsets[cid]
            if len(labelset) == 0:
                continue
            lspec = computeLabelsSpec(labels_to_nodes, nodeset, labelset, data_parameters)
        tmp, ne, nn = addGroup(ti, score(len(edgeset), len(nodeset)), lspec, edgeset, nodeset, labelset=labelset, covered=covered, coveredn=coveredn, nb_e=nb_e, nb_n=nb_n, all_edges=all_edges, graph=graph)
        fo.write(tmp+"\n")
        if fc is not None:
            fc.write("%d\t%d\t%s\t%s\n" % ( len(nodeset), len(labelset),
                                            ", ".join(sorted(nodeset)),
                                            ", ".join(sorted(labelset))))


########################################################
################# SCORING
def score(nb_edges, nb_nodes=None, new_edges=None):
    ### new_edges is not used in this score, only here for compatibility
    if nb_nodes is None: ## in this case provided the actual edges and not just their number 
        nb_nodes = len(getNodeset(nb_edges))
    if nb_edges > 0:
        return nb_nodes/(2.0*nb_edges)
    return float("Inf")

def score1(nb_edges, nb_nodes=None, new_edges=None):
    if nb_nodes is None: ## in this case provided the actual edges and not just their number 
        nb_nodes = len(getNodeset(nb_edges))
    if type(nb_edges) is not int: ## if actual edges provided turn to number
        nb_edges = len(nb_edges)
    if new_edges is None:
        new_edges = nb_edges
    if nb_edges > 0:
        return ((nb_nodes*(nb_nodes-1))/2.0-nb_edges)/new_edges
    return float("Inf")

def density(edgeset, nbN=None):
    if nbN is None:
        nbN = len(getNodeset(edgeset))
    r = (nbN*(nbN-1))-2*len(edgeset)
    return r

def densityR(edgeset, nbN=None):
    if nbN is None:
        nbN = len(getNodeset(edgeset))
    if nbN > 1:
        r = 2.0*len(edgeset)/(nbN*(nbN-1))
    else:
        r = float("Inf")
    return r

def ratio(edgeset, nbN=None):
    if nbN is None:
        nbN = len(getNodeSet(edgeset))
    r = 2.0*len(edgeset)/(nbN*(nbN-1)- 2.0*len(edgeset))
    return r

def computeLabelsSpec(labels_to_nodes, nodeset, labelset, data_parameters):
    if len(labelset) == 0:
        return 0.0
    supp, suppe = getSupportInter(labelset, labels_to_nodes)
    return len(nodeset.intersection(supp))/(1.0*len(nodeset.union(supp)))

def chooseLabels(labels_to_nodes, nodeset, data_parameters):
    candidates = sorted([l for l, supp in labels_to_nodes.items() if len(supp & nodeset) > 0])
    best = (0, None)
    for l in candidates:
        if min(len(nodeset), len(labels_to_nodes[l]))/float(max(len(nodeset), len(labels_to_nodes[l]))) > best[0]:
            t = len(nodeset.intersection(labels_to_nodes[l]))/(1.0*len(nodeset.union(labels_to_nodes[l])))
            if t > best[0]:
                best = (t, l)
    if best[1] is None:
        return [], 0
    top = (best[0], [best[1]], set(labels_to_nodes[best[1]]))
    while top[1] is not None:
        best = top
        top = (best[0], None, None)
        for l in candidates:
            supp = labels_to_nodes[l].intersection(best[2])
            if min(len(nodeset), len(supp))/float(max(len(nodeset), len(supp))) > top[0]:
                t = len(nodeset.intersection(supp))/(1.0*len(nodeset.union(supp)))
                if t > top[0]:
                    top = (t, [l]+best[1], supp)
    return best[1], best[0]


def chooseLabelsTheta(labels_to_nodes, nodeset, theta, feats, data_parameters):
    candidates = [feats[k] for (k, v) in sorted(enumerate(theta[1:]), key=lambda x:x[1]) if v > 0]
    best = [0, [], set(nodeset)]
    while len(candidates) > 0:
        l = candidates.pop()
        suppL = labels_to_nodes.get(l,[])
        if min(len(best[2]), len(suppL))/float(max(len(best[2]), len(suppL))) > best[0]:
            t = len(best[2].intersection(suppL))/(1.0*len(best[2].union(suppL)))
            if t > best[0]:
                best[1].append(l)
                best[0] = t
                best[2].intersection_update(suppL)
            else:
                candidates = []
        else:
            candidates = []
    pdb.set_trace()
    print best
    return best[1], best[0]


def getSupportInter(labelset, labels_to_nodes, labels_to_edges=None):
    cnodes = set(labels_to_nodes.get(labelset[0], set()))
    for l in labelset[1:]:
        cnodes &= labels_to_nodes.get(l, set())

    if labels_to_edges is not None:
        cedges = set(labels_to_edges.get(labelset[0], set()))
        for l in labelset[1:]:
            cedges &= labels_to_edges.get(l, set())
    else:
        cedges = None
    return cnodes, cedges

def getSupportMajority(labelset, nodes_to_labels, edges_to_labels=None, all_edges=None):
    if all_edges is None and edges_to_labels is not None:
        all_edges = set(edges_to_labels.keys())
    lss = set(labelset)
    mini = len(labelset)/2+1
    try:
        cnodes = set([node for (node, ls) in nodes_to_labels.items() if len(lss.intersection(ls)) >= mini])
    except AttributeError:
        pdb.set_trace()
        print "A"
        cnodes = set()
    if all_edges is not None:
        cedges = set([edge for edge in all_edges if edge[0] in cnodes and edge[1] in cnodes])
    else:
        cedges = None
    return cnodes, cedges

def getSupport(stype, labelset, labels_to_nodes=None, labels_to_edges=None, nodes_to_labels=None, edges_to_labels=None, all_edges=None):
    if stype == "maj" and (labels_to_edges is None or len(labelset) > 2):
        return getSupportMajority(labelset, nodes_to_labels, edges_to_labels, all_edges)
    else:
        return getSupportInter(labelset, labels_to_nodes, labels_to_edges)

def getSizeLInter(node, labelset, nodes_to_labels):
    return len(nodes_to_labels.get(node, set()).intersection(labelset))
    
def improveSubgraph(graph, edgeset, covered, scoref=densityR, f=1):
    cdR = scoref(edgeset)
    odR = cdR
    nodes = getNodeset(edgeset)
    neighs = set()
    for node in nodes:
        neighs.update(graph[node].keys())
    added = []
    nedges = set()
    neighs = sorted([(n, nodes & set(graph[n].keys())) for n in neighs - nodes], key= lambda x:len(x[1]))
    while len(neighs) > 0 and len(added) < 0.1*len(nodes):
        n, nds = neighs.pop()
        eds = set([tuple(sorted([n, o])) for o in nds])
        tdR = scoref(edgeset | nedges | eds)
        if f*tdR > f*cdR:
            cdR = tdR
            nedges.update(eds)
            added.append(n)
    #print "Improved with %d nodes from %f to %f" % (len(added), odR, cdR) 
    return nedges, added

########################################################
################# MINING
def densestMaj(nodes_to_labels, labels_to_nodes, all_edges, assigned=None, sizes=None, debug_add=None):
    distsN = [set(nodes_to_labels.keys())]
    onemissE = [[], [], list(all_edges)]
    uncoveredE, coveredE = [], []
    minC = 1
    scores = []
    labelset = []
    cand_labels = labels_to_nodes.keys()
    while True:
        if debug_add is not None:
            debug_add["lbls"] = labelset
        sc, label = bestLabelMaj(cand_labels, labels_to_nodes, minC, distsN, onemissE, coveredE, assigned, sizes, debug_add=debug_add)
        if label is None:
            break
        cand_labels.remove(label)
        labelset.append(label)
        scores.append(sc)
        old_minC = minC
        minC = (len(labelset)+3)/2 
        if debug_add is not None: ## DEBUG ##
            cN = sum([len(c) for c in distsN[old_minC:]])
            cE = len(coveredE)
            N, E = NandELabelMaj(labels_to_nodes[label], old_minC, distsN, onemissE)
            cnodes, cedges = getSupport("maj", labelset, nodes_to_labels=nodes_to_labels, all_edges=all_edges)
            if cN+N != len(cnodes) or cE+E != len(cedges):
                print "TOP %s N:%d+%d=%d E:%d+%d=%d" % (labelset, cN, N, cN+N, cE, E, cE+E) 
                print "VS.", len(cnodes), len(cedges)
                pdb.set_trace()
        distsN, onemissE, coveredE, uncoveredE = updateLabelMaj(labelset, labels_to_nodes, nodes_to_labels,
                                                                old_minC, minC, distsN, onemissE, coveredE, uncoveredE)

    return labelset, scores


def updateLabelMaj(labelset, labels_to_nodes, nodes_to_labels, old_minC, minC, distsN, onemissE, coveredE, uncoveredE):
    ### update list of nodes by intersection size with label set
    up_distsN = [set()]
    for distN in distsN:
        up_distsN[-1].update(distN.difference(labels_to_nodes[labelset[-1]]))
        up_distsN.append(distN.intersection(labels_to_nodes[labelset[-1]]))

    ### update list of edges with one label missing on either or both ends of the edge
    ## minC = old_minC + 0 or 1 
    if minC == old_minC:
        up_coveredE = list(coveredE)
        up_uncoveredE = []
        up_onemissE = [[], [], []]
        retest = uncoveredE
        for i in [0,1]:
            for edge in onemissE[i]:
                if edge[i] in labels_to_nodes[labelset[-1]]:
                    up_coveredE.append(edge)
                else:
                    up_onemissE[i].append(edge)

        for i in [2]:
            for edge in onemissE[i]:
                if edge[0] in labels_to_nodes[labelset[-1]] and edge[1] in labels_to_nodes[labelset[-1]]:
                    up_coveredE.append(edge)
                elif edge[0] in labels_to_nodes[labelset[-1]]:
                    up_onemissE[1].append(edge)
                elif edge[1] in labels_to_nodes[labelset[-1]]:
                    up_onemissE[0].append(edge)
                else:
                    up_onemissE[2].append(edge)

    else:
        up_coveredE = []
        up_uncoveredE = []
        up_onemissE = [[], [], []]
        retest = uncoveredE + coveredE
        for cE in onemissE:
            retest.extend(cE)
        
    for edge in retest:
        c0, c1 = getSizeLInter(edge[0], labelset, nodes_to_labels), getSizeLInter(edge[1], labelset, nodes_to_labels)
        if c0 < minC-1 or c1 < minC-1:
            up_uncoveredE.append(edge)
        elif c0 >= minC and c1 >= minC:
            up_coveredE.append(edge)
        elif c0 == minC-1:
            if c1 == minC-1:
                up_onemissE[2].append(edge)
            else:
                up_onemissE[0].append(edge)
        else:
            up_onemissE[1].append(edge)
                
    return up_distsN, up_onemissE, up_coveredE, up_uncoveredE

def bestLabelMaj(cand_labels, labels_to_nodes, minC, distsN, onemissE, coveredE, assigned=None, sizes=None, debug_add=None):
    top = (float("Inf"), None)
    cN = sum([len(c) for c in distsN[minC:]])
    acE = len(coveredE)
    for label in cand_labels:
        N, E = NandELabelMaj(labels_to_nodes[label], minC, distsN, onemissE, assigned, sizes, cN)
        if assigned is not None:
            cE = len([c for c in coveredE if sizes[assigned.get(c, None)] > N+cN])
        else:
            cE = acE
        if cE+E > 0:
            sc = score(E+cE,N+cN)

            if debug_add is not None:
                cnodes, cedges = getSupport("maj", [label]+debug_add["lbls"], debug_add["ln"], debug_add["le"], nodes_to_labels=debug_add["nl"], edges_to_labels=debug_add["el"])
                ce = 0
                for e in cedges:
                    if sizes[assigned.get(e, None)] > len(cnodes):
                        ce +=1
                if score(ce, len(cnodes)) != sc:
                    N, E = NandELabelMaj(labels_to_nodes[label], minC, distsN, onemissE, assigned, sizes, cN)
                print "COMP SCORES", score(ce, len(cnodes)), sc

            if sc < top[0]:
                top = (sc, label)
    return top

def NandELabelMaj(label_nsupp, minC, distsN, onemissE, assigned=None, sizes=None, cN=0):
    N = len(label_nsupp & distsN[minC-1])
    if assigned is None:
        E  = sum([v_0 in label_nsupp for (v_0, v_1) in onemissE[0]])
        E += sum([v_1 in label_nsupp for (v_0, v_1) in onemissE[1]])
        E += sum([v_0 in label_nsupp and v_1 in label_nsupp for (v_0, v_1) in onemissE[2]])
    else:
        E  = sum([v_0 in label_nsupp for (v_0, v_1) in onemissE[0] if sizes[assigned.get((v_0, v_1), None)] > N+cN])
        E += sum([v_1 in label_nsupp for (v_0, v_1) in onemissE[1] if sizes[assigned.get((v_0, v_1), None)] > N+cN])
        E += sum([v_0 in label_nsupp and v_1 in label_nsupp for (v_0, v_1) in onemissE[2] if sizes[assigned.get((v_0, v_1), None)] > N+cN])
    return N, E

        
########################################################
################# MINING TOOLS

def parseClustersL(graph, nodes_to_labels, candidates, data_parameters, labelsets=None, fo=None, fc=None):
    ##### PARSE CLUSTERS BASED ONLY ON LABELS RECONSTRUCT SUPPORT!!
    min_supp = data_parameters["min_supp"]
    if fo is None:
        fo = sys.stdout

    nodes = set(graph.keys())
    labels_to_nodes = getReversedLabels(nodes_to_labels, graph, data_parameters)
    labels_to_edges, edges_to_labels = getEdgesCands(nodes_to_labels, graph, data_parameters)

    ticE = datetime.datetime.now()
    supps, suppsn, ncands = [], [], []
    while len(labelsets) > 0:
        cand = labelsets.pop()
        ndc, edgc = getSupport(data_parameters["type_support"], cand, labels_to_nodes, labels_to_edges, nodes_to_labels=nodes_to_labels, edges_to_labels=edges_to_labels)
        if len(ndc) > min_supp:
            supps.append(edgc)
            suppsn.append(ndc)
            ncands.append(cand)
    candidates = ncands
    
    cis = range(len(candidates))
    cis.sort(key= lambda x:(len(suppsn[x]),candidates[x]))
    covered = {}
    coveredn = {}
    nb_n = float(len(graph))
    nb_e = float(len(edges_to_labels))
    all_edges = set(edges_to_labels.keys())
    for ti, cid in enumerate(cis):
        nodeset = suppsn[cid]
        edgeset = supps[cid]
        labelset = candidates[cid]
        lspec = computeLabelsSpec(labels_to_nodes, nodeset, labelset, data_parameters)
        tmp, ne, nn = addGroup(ti, score(len(edgeset), len(nodeset)), lspec, edgeset, nodeset, labelset=labelset, covered=covered, coveredn=coveredn, nb_e=nb_e, nb_n=nb_n, all_edges=all_edges, graph=graph)
        fo.write(tmp+"\n")
        if fc is not None:
            fc.write("%d\t%d\t%s\t%s\n" % ( len(nodeset), len(labelset),
                                            ", ".join(sorted(nodeset)),
                                            ", ".join(sorted(labelset))))

def selectGreedy(graph, nodes_to_labels, candidates, data_parameters, fo=None, fc=None):
    min_supp = data_parameters["min_supp"]
    if fo is None:
        fo = sys.stdout

    nodes = set(graph.keys())
    labels_to_nodes = getReversedLabels(nodes_to_labels, graph, data_parameters)
    labels_to_edges, edges_to_labels = getEdgesCands(nodes_to_labels, graph, data_parameters)

    ticE = datetime.datetime.now()
    if type(candidates) == dict:
        candidates, supps = zip(*candidates.items())
        suppsn = []
        while len(suppsn) < len(candidates):
            nds = getNodeset(supps[len(suppsn)])
            if len(nds) > min_supp:
                suppsn.append(nds)
            else:
                supps.pop(len(suppsn))
                candidates.pop(len(suppsn))
                
    else:     ### generate supports for candidates if not provided
        supps, suppsn, ncands = [], [], []
        while len(candidates) > 0:
            cand = candidates.pop()
            ndc, edgc = getSupport(data_parameters["type_support"], cand, labels_to_nodes, labels_to_edges, nodes_to_labels=nodes_to_labels, edges_to_labels=edges_to_labels)

            if len(ndc) > min_supp:
                supps.append(edgc)
                suppsn.append(ndc)
                ncands.append(cand)
        candidates = ncands
        
    ### start greedy selection
    cis = range(len(candidates))
    cis.sort(key= lambda x:(len(suppsn[x]),candidates[x]))
    nb_e = float(len(edges_to_labels))
    nb_n = float(len(nodes))
    all_edges = set(edges_to_labels.keys())
    covered = {}
    coveredn = {}
    resultsc = 0
    top = (1, 1, 1, 1)
    while top[1] is not None:
        top = (float("Inf"), None, None)
        for ci in cis:
            if len(suppsn[ci].difference(coveredn.keys())) > min_supp: # and r < 1:
                r = score(len(supps[ci].difference(covered)), len(suppsn[ci]), len(supps[ci]))
                if r < top[0]: # and r != 1:
                    top = (r, ci)

        if top[1] is not None:
            lspec = computeLabelsSpec(labels_to_nodes, suppsn[top[1]], candidates[top[1]], data_parameters)
            tmp, ne, nn = addGroup(resultsc, top[0], lspec, supps[top[1]], suppsn[top[1]], labelset=candidates[top[1]], covered=covered, coveredn=coveredn, nb_e=nb_e, nb_n=nb_n, all_edges=all_edges, graph=graph)
            resultsc +=1
            fo.write(tmp+"\n")
            if fc is not None:
                fc.write("%d\t%d\t%s\t%s\n" % ( len(suppsn[top[1]]), len(candidates[top[1]]),
                                                ", ".join(sorted(suppsn[top[1]])),
                                                ", ".join(sorted(candidates[top[1]]))))
            cis.remove(ci)
    print "Done in %s" % (datetime.datetime.now() - ticE)

def getCandidatesFromLO(graph, labels_to_nodes, labels_to_edges, lo, data_parameters,
                        nodes_to_labels=None, edges_to_labels=None):
    start = 0
    mul = 0
    candidates = []
    while start < len(lo):
        lset = set([lo[start]])
        nsupp = labels_to_nodes[lo[start]]
        esupp = labels_to_edges[lo[start]]
        #if score(len(esupp), len(nsupp)) < 3:
        candidates.append({"nodeset":set(nsupp), "edgeset": set(esupp), "labelset": set(lset)})
        end = start+1
        while end < len(lo):
            esupp &= labels_to_edges[lo[end]]
            if len(esupp) < 2:
                end = len(lo)
            else:
                nsupp &= labels_to_nodes[lo[end]]
                lset.add(lo[end])
                #if score(len(esupp), len(nsupp)) < 3:
                mul += 1
                if data_parameters["type_support"] != "int":
                    nodesetB, edgesetB = getSupport(data_parameters["type_support"], list(lset), labels_to_nodes, labels_to_edges,
                                                    nodes_to_labels=nodes_to_labels, edges_to_labels=edges_to_labels)
                    nsupp = nodesetB
                    esupp = edgesetB
                # if nodesetB != nsupp or edgesetB != esupp:
                #     pdb.set_trace()
                #     print len(nsupp)
                candidates.append({"nodeset":set(nsupp), "edgeset": set(esupp), "labelset": set(lset)})
                # candidates.append({"nodeset":set(nodesetB), "edgeset": set(edgesetB), "labelset": set(lset)})
            end += 1
        start +=1
    # print "Returning %d candidates, %d multiple, out of %d labels" % (len(candidates), mul, len(lo))
    return candidates

def getLabelsOrdering(graph, labels_to_nodes, data_parameters):
    dM, labels = makeLabelDistMatrix(labels_to_nodes, data_parameters)
    if dM is not None:
        try:
            fiedo = getFiedlerOrd(dM)
        except Exception:
            fiedo = range(len(labels))
        return [labels[i] for i in fiedo], [-dM[fiedo[i],fiedo[i+1]] for i in range(len(fiedo)-1)]
    return labels, [0 for i in range(len(labels)-1)]

def makeLabelDistMatrix(labels_to_items, data_parameters):
    labels = sorted(labels_to_items.keys())
    vij = []
    for lAi, lA in enumerate(labels):
        vij += [(len(labels_to_items[lA] & labels_to_items[lB])/float(len(labels_to_items[lA] | labels_to_items[lB])), lAi, lBi)
               for lBi, lB in enumerate(labels[:lAi]) if len(labels_to_items[lA] & labels_to_items[lB]) > 0]
    if len(vij) > 0:
        V,I,J = zip(*vij)
        dM = coo_matrix((-numpy.array(V+V), (I+J,J+I)), shape=(len(labels),len(labels)) ).tocsr()
        return dM + spdiags(-dM.sum(0), [0], len(labels), len(labels), "csr"), labels
    else:
        return None, labels

def getFiedlerOrd(dM):
    eval, evect = eigsh(dM, 2)
    return numpy.argsort(evect[:,1])

def mineTriangleNode(node, graph, nodes_to_labels, edges_to_labels, labels_to_nodes, labels_to_edges, data_parameters, seen_labels=None):
    neighs_one = set(graph[node].keys())
    if len(neighs_one) == 1:
        return
    neighs_close = set([n for n in set(itertools.chain(*[graph[n].keys() for n in neighs_one])).difference(neighs_one)
                        if len(neighs_one.intersection(graph[n].keys())) > len(graph[n].keys())*data_parameters["tri_neighf"]])
    neighs_close.update(neighs_one)
    # inter_labels = set(nodes_to_labels[node]).intersection(*[nodes_to_labels[n] for n in neighs_close])
    neighs_close.add(node)
    labelset, lspec = chooseLabels(labels_to_nodes, neighs_close, data_parameters)
    labelset = tuple(labelset)
    # print node, len(neighs_close), inter_labels, labelset
    # pdb.set_trace()
    if len(labelset) > 0 and seen_labels is not None and labelset not in seen_labels:
        seen_labels.add(labelset)
        nodesetB, edgesetB = getSupport(data_parameters["type_support"], labelset, labels_to_nodes, labels_to_edges,
                                        nodes_to_labels=nodes_to_labels, edges_to_labels=edges_to_labels)
        if len(edgesetB) > 0:
            return {"labelset": labelset, "nodeset": nodesetB, "edgeset": edgesetB}

### greedy selection of candidates {"labelset", "nodeset", "edgeset"}, possibly stealing edges (CK)
def selectCandidates(candidates, graph, data_parameters, external, ticE):
    ks = data_parameters["ks"]
    unassigned = len(graph)+10
    edges_nbassi = {}
    assE = {}
    for n in graph.keys():
        for nei in graph[n].keys():
            assE[(nei,n)] = (unassigned, None)
            if nei < n:
                edges_nbassi[(nei,n)] = 1
    nb_n = float(len(graph))
    nb_e = float(len(edges_nbassi))
    all_edges = set(edges_nbassi.keys())

    stop = False
    selected = []
    ti = -1
    while len(candidates) > 0:
        select = sorted([(score(len([e for e in candidate["edgeset"] if assE[e][0] > len(candidate["nodeset"])]),
                                len(candidate["nodeset"])), ci) for ci, candidate in enumerate(candidates)])[0]
        if select[0] < float('inf'):
            ti += 1
            selected.append(candidates.pop(select[1]))
            for e in selected[-1]["edgeset"]:
                if assE[e][0] > len(selected[-1]["nodeset"]):
                    assE[e] = (len(selected[-1]["nodeset"]), len(selected)-1)
        else:
            candidates = []

        if ti+1 >= max(ks):
            candidates = []

        if ti+1 in ks or stop:
            for i in range(ti+1):
                selected[i]["edgeset"] = set()
            for edge, v in assE.items():
                if v[1] != None:
                    selected[v[1]]["edgeset"].add(edge)
            printSelected(selected, graph, data_parameters, external, ticE, nb_n, nb_e, all_edges)

def printSelected(selected, graph, data_parameters, external, ticE, nb_n, nb_e, all_edges):
    idsort = sorted(range(len(selected)), key=lambda x: len(selected[x]["nodeset"]))

    covered = {}
    coveredn = {}

    fo, fe = initResOut(data_parameters, external, len(selected))

    for ti, tt in enumerate(idsort):
        tmp, ne, nn = addGroup(ti, score(len(selected[tt]["edgeset"]), len(selected[tt]["nodeset"])), 1, selected[tt]["edgeset"], selected[tt]["nodeset"], labelset=selected[tt]["labelset"], covered=covered, coveredn=coveredn, nb_e=nb_e, nb_n=nb_n, all_edges=all_edges, graph=graph)
        fo.write(tmp+"\n")
        if fe is not None:
            # fe.write(" ".join(sorted([getNodeId(n, graph, data_parameters) for n in nodesets[tt]])) + "\n")
            fe.write("%d\t%d\t%s\t%s\n" % ( len(selected[tt]["nodeset"]), len(selected[tt]["labelset"]),
                                           " ".join(sorted(selected[tt]["nodeset"])),
                                           " ".join(sorted(selected[tt]["labelset"]))))

    fo.write("Done in %s\n" % (datetime.datetime.now() - ticE))
    if fe is not None:
        fo.close()
        fe.close()

def makeRndOpinions(graph, parameters, nbo):
    #### HERE: TRY OTHER TOO
    typ = parameters.get("oopityp", "u")
    if typ == "u":
        return numpy.random.rand(len(graph), nbo)
    return numpy.random.rand(len(graph), nbo)


####################################################
###### DRAWING
def getNbEdges(graph):
    return sum([len(neighs) for n, neighs in graph.items()])/2.0

def getEdgeset(nodes, graph):
    edgeset = set()
    for nodeA in nodes:
        for nodeB in graph.get(nodeA, {}).keys():
            if nodeA < nodeB and nodeB in nodes:
                edgeset.add((nodeA, nodeB))
    return edgeset

def getFront(nodesetA, nodesetB, graph, reverse=False):
    points = []
    if reverse:
        sourceset = nodesetB
        targetset = nodesetA
    else:
        sourceset = nodesetA
        targetset = nodesetB
    
    N = float(len(sourceset))
    E = len(getEdgeset(sourceset, graph))

    nodes_add = targetset.difference(sourceset)
    nodes_rem = sourceset.difference(targetset)
    
    lparts = (len(targetset.intersection(sourceset)), len(nodes_add), len(nodes_rem)) 
    if reverse:
        nbs = {"inter": N,
               "union": N }
        nbs_from = "inter"
        nbs_to = "union"
    else:
        nbs = {"inter": N - len(nodes_rem),
               "union": N + len(nodes_add)}
        nbs_from = "union"
        nbs_to = "inter"

    degrees = {}
    for node in (nodes_add | nodes_rem):
        degrees[node] = len(sourceset.intersection(graph[node].keys()))

    if E > 0:
        sc = N/(2*E)
    else:
        sc = DEF
    points.append((sc, nbs["inter"] / nbs["union"], nbs["union"]/10.0))
    

    while len(degrees) > 0:
        top = sorted([((N+1)/(2*(E+degrees[node])), True, node) for node in nodes_add if E+degrees[node] > 0] + [(N/(2*(E-degrees[node])), False, node) for node in nodes_rem if E-degrees[node] > 0]+[(float("Inf"), node in nodes_add, node) for node in random.sample(degrees.keys(), 1)])[0]
        
        d = degrees.pop(top[2])
        if top[1]:
            nodes_add.remove(top[2])
            nbs[nbs_to] += 1
            off = 1
        else:
            nodes_rem.remove(top[2])
            nbs[nbs_from] -= 1
            off = -1
            
        for node in degrees.keys():
            if top[2] in graph.get(node, {}).keys():
                degrees[node] += off
        N += off
        E += off*d

        if E > 0:
            sc = N/(2*E)
        else:
            sc = DEF
        points.append((sc, nbs["inter"] / nbs["union"], nbs["union"]/10.0))
    return points, lparts


def makeColor(color_base, factors, ratio):
    colorv = list(color_base)
    for i in range(3):
        if factors[i] == 1:
            colorv[i] *= ratio 
        elif factors[i] == -1:
            colorv[i] =  1-(colorv[i]*ratio) 
    return tuple(colorv)

def graphTo2D(graph, nodes_to_locations, data_parameters):
    store_utils = {}
    if data_parameters["ndlb_typ"] == "coo":
        dims = range(len(nodes_to_locations.values()[0]))
        if len(dims) == 2:
            for n, coo in nodes_to_locations.items():
                store_utils[n] = [coo[d] for d in dims]
    if data_parameters["ndlb_typ"] == "lbl":
        from scipy.sparse import coo_matrix
        from sklearn.manifold import Isomap
        map_lbls = {}
        map_nodes = {}
        rowids = []
        colids = []
        for ni, (n, lbls) in enumerate(nodes_to_locations.items()):
            map_nodes[n] = ni
            for lbl in lbls:
                if lbl not in map_lbls:
                    map_lbls[lbl] = len(map_lbls)
                colids.append(ni)
                rowids.append(map_lbls[lbl])

        vals = numpy.ones(len(rowids))
        data = coo_matrix((vals,(colids, rowids)))
        projected_data = Isomap(n_neighbors=10, n_components=2).fit_transform(data.todense())
        for n, ni in map_nodes.items():
            store_utils[n] = (projected_data[ni, 0], projected_data[ni, 1])
        # from sklearn import manifold: # datasets, decomposition, ensemble, random_projection)

    return store_utils


#######################################
### TEAM STUFF
#######################################


def logAndDetails(graph, nopis, comm, data_parameters, fo=None, more={}):
    details = {}
    if comm is not None:
        comm_org = comm
        comm = [c for c in comm_org if c < nopis.shape[0]]
        ss = nopis[comm]
        adj, dia, _ = graph_to_mats(graph, nopis, comm, data_parameters)
        
        more["connect_graph"] = True
        if len(comm_org) > len(comm): # True: #TEST_CONN and len(comm_org) == len(comm):
            print "Check"
            G = nkx.Graph(adj)
            if not nkx.is_connected(G):
                return None
                more["connect_graph"] = False
                # print "The graph is not connected!!!"
                # pdb.set_trace()
                # nkx.draw_networkx(G) #, node_color='grey', pos=pos, node_size=1, edge_color='grey', with_labels=False)
        if adj.sum() == 0: ### nodes are not connected...
            opd = numpy.zeros((1, ss.shape[1]))
            zz = ss
            Ui, Uc = (0., 0.)
        else:
            
            numpy.prod(adj.shape)
            opd = noavg_opd(ss, adj)
            zz = get_zzs_comm(adj, dia, ss, data_parameters)
            Ui, Uc = part_tension(ss, zz, adj, dia)
        U = Ui+Uc
        details = {"mats": (adj, dia), "zz": zz, "ss": ss, "U": U, "i": Ui, "c": Uc,
                   "nbN": numpy.sum(dia>0), "nbE": numpy.sum(adj>0)/2.}
        details.update(get_stats_opd(opd))
        details["highest_ediff"] = max_edges(ss, adj)
        details["external"] = data_parameters.get("external", "---")
        print "U=%f Ui=%f Uc=%f N=%d E=%d" % (U,  Ui, Uc, numpy.sum(dia>0), numpy.sum(adj>0)/2)
        if fo is not None:
            fo.write("U=%f Ui=%f Uc=%f N=%d E=%d\n" % (U,  Ui, Uc, numpy.sum(dia>0), numpy.sum(adj>0)/2))
            fo.write("opd_max=%(opd_max)f opd_mean=%(opd_mean)f opd_median=%(opd_median)f opd_tot=%(opd_tot)f opd_var=%(opd_var)f opd_len=%(opd_len)f\n" % details)
            if "map_iton" in data_parameters:
                fo.write("\t".join(["%s" % data_parameters["map_iton"][n] for n in comm])+"\n")
            fo.write("\t".join(["%d" % n for n in comm])+"\n")
            fo.write("ss="+numpy.array2string(ss, max_line_width=10**9, formatter={'float_kind':lambda x: "%f" % x})+"\n")
            fo.write("zz="+numpy.array2string(zz, max_line_width=10**9, formatter={'float_kind':lambda x: "%f" % x})+"\n")
            # fo.write("\t".join([numpy.array2string(ss[i,:]) for i in range(zz.shape[0])])+"\n")
            # fo.write("\t".join(["%s" % zz[i,:] for i in range(zz.shape[0])])+"\n")

        if "bckb" in more:
            commb = [c for c in more["bckb"].keys() if c < nopis.shape[0]]
            ssb = nopis[commb]
            adjb, diab, _ = graph_to_mats(more["bckb"], nopis, commb, data_parameters)
            if adjb.sum() == 0: ### nodes are not connected...
                zzb = ssb
                Uib, Ucb = (0., 0.)
            else:
                zzb = get_zzs_comm(adjb, diab, ssb, data_parameters)
                Uib, Ucb = part_tension(ssb, zzb, adjb, diab)
            Ub = Uib+Ucb
            details.update({"zz_bckb": zzb, "U_bckb": Ub, "i_bckb": Uib, "c_bckb": Ucb,
                   "nbN_bckb": numpy.sum(diab>0), "nbE_bckb": numpy.sum(adjb>0)/2.})
            if fo is not None:
                fo.write("BCKB U=%f Ui=%f Uc=%f N=%d E=%d\n" % (Ub,  Uib, Ucb, numpy.sum(diab>0), numpy.sum(adjb>0)/2))

            more["connect_bckb"] = True
            if TEST_CONN and len(comm_org) == len(comm):
                G = nkx.Graph(adjb)
                if not nkx.is_connected(G):
                    more["connect_bckb"] = False
                    # print "The bckb graph is not connected!!!"
                    # pdb.set_trace()

    if fo is not None:
        fo.write("*\t"+ ("\t".join(["%s=%s" % (k,v) for k,v in more.items() if k != "bckb"]))+"\n")
    details.update(more)
    if "runtime" in details:
        details["rtime"] = details["runtime"].total_seconds() 
    return details


def plot_comm(graph, nopis, details, comm, data_parameters, title_pre=""):
    if "pos" in data_parameters:
        pos = data_parameters["pos"]
    else:
        pos = generate_positions(graph)
    if "peaks" in data_parameters:
        tops, lows = data_parameters["peaks"]
    else:
        tops, lows = findPeaks(graph, nopis)
        data_parameters["peaks"] = (tops, lows) 
    ss = nopis

    N = len(graph)
    G = nkx.Graph()
    G.add_edges_from([(n,n+N) for n in comm])
    # G.add_nodes_from([n for n in range(N)]+[N+n for n in range(N)])
    ssc, zzc = details["ss"], details["zz"]

    comm_edges = []
    scomm = set(comm)
    for n in comm:
        comm_edges.extend([(n,neigh) for neigh in scomm.intersection(graph[n].keys())])
    H = nkx.Graph()
    H.add_edges_from(comm_edges)
    nkx.draw_networkx_nodes(G, nodelist=comm, node_color=zzc, pos=pos, node_size=EZ, with_labels=False, cmap=plt.cm.Blues, vmin=0, vmax=1)

    nkx.draw_networkx(G, nodelist=[n for n in comm] + [N+n for n in comm], node_color='grey', pos=pos, node_size=1, edge_color='grey', with_labels=False)

    nkx.draw_networkx_nodes(G, nodelist=[N+n for n in comm], node_color=ssc, pos=pos, node_size=IZ, with_labels=False, cmap=plt.cm.Blues, vmin=0, vmax=1)
    # pcc = nkx.draw_networkx_nodes(G, nodelist=[N+t for t in tops], pos=pos, node_size=1, with_labels=False)
    # pcc.set_edgecolor("red")
    # pcc = nkx.draw_networkx_nodes(G, nodelist=[N+t for t in lows], pos=pos, node_size=1, with_labels=False)
    # pcc.set_edgecolor("green")
    nkx.draw_networkx_edges(H, node_color=zzc, edge_color='black', pos=pos)
    nkx.draw_networkx_labels(H, pos=pos, font_color='red', font_size=10)

    plt.xlabel("U=%f E=%d\ni=%f + c=%f" % (details["U"], len(H.edges()), details["i"], details["c"]))
    plt.title("%s k=%d" % (title_pre, data_parameters["k"]))
    # for nd in comm:
    #     # try:
    #     #     plt.text(pos[nd+N][0], pos[nd+N][1], data_parameters['map_iton'][nd])
    #     # except ValueError:
    #     plt.text((pos[nd+N][0]+pos[nd][0])/2, (pos[nd+N][1]+pos[nd][1])/2, nd)
    #     print nd,"-->", data_parameters['map_iton'][nd]
    plt.xticks([])
    plt.yticks([])
    plt.xlim(-1.5,1.5)
    plt.ylim(-1.5,1.5)
    plt.draw()

def plot_comm_big(graph, nopis, details, comm, data_parameters, title_pre=""):
    if "mats" in data_parameters:
        adj, dia, _ = data_parameters["mats"]
    else:
        adj, dia, _ = graph_to_mats(graph, nopis, None, data_parameters)[:2]
    if "pos" in data_parameters:
        pos = data_parameters["pos"]
    else:
        pos = generate_positions(graph, adj)
    if "peaks" in data_parameters:
        tops, lows = data_parameters["peaks"]
    else:
        tops, lows = findPeaks(graph, nopis)
        data_parameters["peaks"] = (tops, lows) 
    if "G" in data_parameters:
        G = data_parameters["G"]
    else:
        fad = numpy.vstack([numpy.hstack([adjG, diaG]), numpy.hstack([numpy.zeros(adjG.shape), numpy.eye(adjG.shape[0])])])
        G = nkx.Graph(fad)

    ss = nopis

    N = len(graph)
    ssc, zzc = details["ss"], details["zz"]
    nkx.draw_networkx_nodes(G, nodelist=comm, node_color=zzc, pos=pos, node_size=EZ, with_labels=True, cmap=plt.cm.Blues, vmin=0, vmax=1)

    nkx.draw_networkx_edges(G, node_color='grey', pos=pos, edge_color='grey')
    nkx.draw_networkx_nodes(G, nodelist=[N+n for n in range(N)], node_color=ss, pos=pos, node_size=IZ, with_labels=False, cmap=plt.cm.Blues, vmin=0, vmax=1)
    # pcc = nkx.draw_networkx_nodes(G, nodelist=[N+t for t in tops], pos=pos, node_size=1, with_labels=False)
    # pcc.set_edgecolor("red")
    # pcc = nkx.draw_networkx_nodes(G, nodelist=[N+t for t in lows], pos=pos, node_size=1, with_labels=False)
    # pcc.set_edgecolor("green")

    H = G.subgraph(comm)
    # nkx.draw_networkx(H, nodelist=comm, node_color=zzc, edge_color='black', pos=pos, node_size=EZ, with_labels=True, cmap=plt.cm.Blues, vmin=0, vmax=1)
    nkx.draw_networkx_edges(H, node_color=zzc, edge_color='black', pos=pos)
    posl = dict([(nk, (pos[nk][0], pos[nk][1]+0.01)) for nk in comm])
    nkx.draw_networkx_labels(H, pos=posl, font_color='red', font_size=10)

    if "ans" in data_parameters and len(data_parameters["ans"][data_parameters["opid"]]) > 0:
        anchors = [data_parameters["map_ntoi"][ii] for ii in data_parameters["ans"][data_parameters["opid"]]]
        pcc = nkx.draw_networkx_nodes(G, nodelist=anchors, pos=pos, node_size=3, with_labels=False)
        pcc.set_edgecolor("green")

    plt.xlabel("U=%f E=%d\ni=%f + c=%f" % (details["U"], len(H.edges()), details["i"], details["c"]))
    plt.title("%s k=%d" % (title_pre, data_parameters["k"]))
    # for nd in comm:
    #     # try:
    #     #     plt.text(pos[nd+N][0], pos[nd+N][1], data_parameters['map_iton'][nd])
    #     # except ValueError:
    #     plt.text((pos[nd+N][0]+pos[nd][0])/2, (pos[nd+N][1]+pos[nd][1])/2, nd)
    #     print nd,"-->", data_parameters['map_iton'][nd]
    plt.xticks([])
    plt.yticks([])
    plt.xlim(-1.5,1.5)
    plt.ylim(-1.5,1.5)

    plt.draw()

def plot_edges_big(graph, nopis, details, edges, data_parameters, title_pre=""):
    if "mats" in data_parameters:
        adj, dia = data_parameters["mats"]
    else:
        adj, dia, _ = graph_to_mats(graph, nopis, None, data_parameters)
    if "pos" in data_parameters:
        pos = data_parameters["pos"]
    else:
        pos = generate_positions(graph, adj)
    if "peaks" in data_parameters:
        tops, lows = data_parameters["peaks"]
    else:
        tops, lows = findPeaks(graph, nopis)
        data_parameters["peaks"] = (tops, lows) 
    ss = nopis

    N = len(graph)
    fad = numpy.vstack([numpy.hstack([adj, dia]), numpy.hstack([numpy.zeros(adj.shape), numpy.eye(adj.shape[0])])])
    G = nkx.Graph(fad)
    ssc, zzc = details["ss"], details["zz"]
    nkx.draw_networkx_nodes(G, nodelist=[n for n in range(N)] + [N+n for n in range(N)], node_color=numpy.concatenate([zzc,ssc]), pos=pos, node_size=IZ, with_labels=False, cmap=plt.cm.Blues, vmin=0, vmax=1)

    nkx.draw_networkx_edges(G, node_color='grey', pos=pos, edge_color='grey')
    nkx.draw_networkx_edges(G, edgelist=edges, pos=pos, edge_color='black')

    # if "ans" in data_parameters and len(data_parameters["ans"][data_parameters["opid"]]) > 0:
    #     anchors = [data_parameters["map_ntoi"][ii] for ii in data_parameters["ans"][data_parameters["opid"]]]
    #     pcc = nkx.draw_networkx_nodes(G, nodelist=anchors, pos=pos, node_size=3, with_labels=False)
    #     pcc.set_edgecolor("green")

    plt.xlabel("U=%f\n%f + %f" % (details["tension"], details["inner"], details["cross"]))
    plt.title("%s" % (title_pre))
    # for nd in comm:
    #     # try:
    #     #     plt.text(pos[nd+N][0], pos[nd+N][1], data_parameters['map_iton'][nd])
    #     # except ValueError:
    #     plt.text((pos[nd+N][0]+pos[nd][0])/2, (pos[nd+N][1]+pos[nd][1])/2, nd)
    #     print nd,"-->", data_parameters['map_iton'][nd]
    plt.xticks([])
    plt.yticks([])
    plt.xlim(-1.5,1.5)
    plt.ylim(-1.5,1.5)

    plt.draw()

if __name__ == "__main__":
    generateGraph()



            # ndlb_dict, typ = readLabels(data_parameters['data_rep']+ bf[1], graph=graph, min_cl=data_parameters["min_count"])
            # getCoocLabels(ndlb_dict)

            # lbls, nnbs = zip(*[(p,len(nds)) for (p, nds) in reverseNdlb(ndlb_dict).items()])
            # nnbs = numpy.array(nnbs, dtype=float)
            # llll = sorted(enumerate(lbls), key=lambda x: nnbs[x[0]])
            # data_parameters["lbl_filterset"] = set(numpy.random.choice(lbls, size=10, replace=False, p=nnbs/numpy.sum(nnbs)))
            # pdb.set_trace()

            # [(1921, 'adaptiv'), (2089, 'image'), (4488, 'rule'), (14090, 'databas'), (12963, 'discovery'), (17388, 'local'), (3872, 'optimiz'), (2853, 'set'), (10322, 'query'), (4621, 'study'), (8899, 'evalu'), (476, 'tree'), (17786, 'automat'), (3583, 'process'), (1012, 'support'), (5828, 'multipl'), (11742, 'distribut'), (9061, 'new'), (12179, 'larg'), (5620, 'graph'), (5899, 'dynam'), (4255, 'web'), (10296, 'predict'), (2714, 'comput'), (11329, 'search'), (16011, 'framework'), (12988, 'pattern'), (2384, 'select'), (1116, 'time'), (10002, 'featur'), (12824, 'structur'), (7601, 'informat'), (13089, 'multy'), (5708, 'gener'), (2332, 'method'), (1667, 'detect'), (2093, 'efficy'), (8616, 'applicat'), (10001, 'classif'), (4856, 'network'), (2426, 'learn'), (6159, 'analysy'), (3823, 'system'), (10859, 'cluster'), (14100, 'approach'), (3330, 'algorithm'), (9453, 'mine'), (15389, 'model'), (8272, 'data'), (7123, 'base')]


            # # [(58, 'improv'), (58, 'path'), (58, 'polynomy'), (58, 'power'), (58, 'theory'), (58, 'toward'), (60, 'larg'), (60, 'match'), (60, 'query'), (61, 'fast'), (61, 'local'), (61, 'new'), (63, 'commun'), (63, 'multy'), (65, 'set'), (67, 'random'), (68, 'function'), (68, 'game'), (69, 'two'), (70, 'search'), (71, 'program'), (72, 'dynam'), (72, 'linear'), (73, 'informat'), (75, 'analysy'), (75, 'design'), (75, 'extend'), (76, 'abstract'), (77, 'optimiz'), (82, 'approach'), (82, 'distribut'), (82, 'tree'), (83, 'data'), (83, 'structur'), (85, 'optimal'), (87, 'approxim'), (90, 'time'), (93, 'system'), (94, 'applicat'), (94, 'bound'), (94, 'graph'), (95, 'efficy'), (97, 'gener'), (98, 'base'), (98, 'model'), (102, 'complex'), (106, 'problem'), (108, 'comput'), (111, 'network'), (126, 'algorithm')]

            # data_parameters["lbl_filterset"] = set(["complex", "problem", "bound", "approxim", "fast", "search"])
            # print data_parameters["lbl_filterset"]
            # nopis = numpy.zeros(len(graph))
            # nbfl = float(len(data_parameters["lbl_filterset"]))
            # for k, wds in ndlb_dict.items():
            #     if k in data_parameters["map_ntoi"]:
            #         # nopis[data_parameters["map_ntoi"][k]] = float(numpy.sum([wds.get(tf, 0.0) for tf in data_parameters["lbl_filterset"]])) / numpy.sum(wds.values())
            #         # nopis[data_parameters["map_ntoi"][k]] = float(len(data_parameters["lbl_filterset"].intersection(wds.keys()))) / len(wds)
            #         nopis[data_parameters["map_ntoi"][k]] = len(data_parameters["lbl_filterset"].intersection(wds.keys())) / nbfl

# def getCoocLabels(ndlb_dict):
#     lbls, nnbs = zip(*[(p,len(nds)) for (p, nds) in reverseNdlb(ndlb_dict).items() if len(nds) > 2])
#     llll = sorted(enumerate(lbls), key=lambda x: nnbs[x[0]], reverse=True)
#     Nf = 100
#     INFN = len(ndlb_dict)
#     dict_wds = dict([(wd[1], ii) for (ii, wd) in enumerate(llll)])
#     mat_cooc = dok_matrix((len(llll), Nf), dtype=numpy.int8)

#     from sklearn import decomposition
#     decomposition.RandomizedPCA(n_components=10).fit_transform()
#     decomposition.RandomizedPCA.fit_transform(X)

#     cols, rows = ([], [])
#     for k, wdks in ndlb_dict.items():
#         wds = [dict_wds[w] for w in wdks.keys() if w in dict_wds]
#         topwds = [i for i in wds if i < Nf]
#         if len(topwds) > 0:
#             for i in topwds:
#                 for j in wds:
#                     try:
#                         mat_cooc[j,i] += 1
#                     except IndexError:
#                         print j,i, mat_cooc.shape
#                         pdb.set_trace()
#     pdb.set_trace()
#     print mat_cooc.shape

#### MORE TEAM STUFF
def proj_mat(adj, nopis=None, Qub=None, thres=0, typ=None, nbdim=1):
    L = get_weighted_L(adj, nopis, Qub, thres, typ)
    ks = [v+1 for v in range(nbdim)]
    return fiedler_val(L, ks)

def get_neighbors(center, dists, k):
    ttt = numpy.argsort(dists[center,:])
    return ttt[:k]

def get_weighted_L(adj, nopis=None, Qub=None, thres=0, typ=None):
    diff_mat = numpy.abs(diffs(nopis, nopis))
    if thres < 0:
        thres = numpy.percentile(diff_mat[adj>0], -thres)
    if thres == 0:
        mat = 1.*(adj>0)
    else:
        mat = 1.*(diff_mat<thres)*(adj>0)

    if typ == "exp_opi":
        mat *= numpy.exp(-diff_mat)
    elif typ == "lin_opi":
        mat  *= 1./(0.0001+diff_mat)
    ## elif typ == "adj":
    ##    mat *= adj
    L = -mat + numpy.diag(numpy.sum(mat,1))
    return L


def fiedler_order_all(adj, nopis=None, Qub=None, thres=0, typ=None):
    L = get_weighted_L(adj, nopis, Qub, thres, typ)
    tord, tval = fiedler_ord(L)
    return tord

def fiedler_order_wei(graph, nopis, adj, Qub, thres=0.10):
    diff_mat = numpy.abs(diffs(nopis, nopis))
    # mat = numpy.ones_like(adj)*(adj>0) #numpy.exp(-(diff_mat*Qub))*
    # mat = numpy.exp(-numpy.divide(diff_mat,Qub))*(adj>0)
    # mat = numpy.exp(-diff_mat)*(adj>0)
    # import matplotlib.pyplot as plt
    # plt.hist(mat[mat!=0],100)
    # # plt.plot(xxA, "b")
    # plt.show()
    if thres < 0:
        thres = numpy.percentile(diff_mat[adj>0], -thres)
    # mat = (diff_mat<thres)*(adj>0) * 1./(0.0001+diff_mat)
    mat = (diff_mat<thres)*(adj>0) # * 1./(0.0001+diff_mat)
    # import matplotlib.pyplot as plt
    # plt.hist(mat[mat!=0],100)
    # # plt.plot(xxA, "b")
    # plt.show()
         
    # mat = 1./(0.00001+diff_mat)*(adj>0)
    # mat += 0.1*numpy.exp(-Qub*diff_mat)*(adj==0)
    # mat = 1.*(adj>0) #-5.*(adj==0)
    L = -mat + numpy.diag(numpy.sum(mat,1))    
    tord, tval = fiedler_ord(L)
    # # # mat[range(mat.shape[0]), range(mat.shape[0])] = 0    
    # L = -mat + numpy.diag(numpy.sum(mat,1))
    # tordA, tvalA = fiedler_ord(L)
    # try:
    #     tordA, tvalA = fiedler_ord(L)
    # except Exception as e:
    #     pdb.set_trace()
    #     tordA = numpy.random.rand(L.shape[0])
    # if numpy.sum(tordA != tord) and numpy.sum(tordA[-1::-1] != tord):
    #     pdb.set_trace()
    # xx = (tval-min(tval))/(max(tval)-min(tval))
    # xxA = (tvalA-min(tvalA))/(max(tvalA)-min(tvalA))
    # print (numpy.mean(xx) - numpy.mean(xxA))**2
    # if (numpy.mean(xx) - numpy.mean(xxA))**2 > 0.1:
    #     pdb.set_trace()
    #     import matplotlib.pyplot as plt
    #     plt.plot(xx, "r")
    #     plt.plot(xxA, "b")
    #     plt.show()
    return tord

def fiedler_order_opi(graph, nopis, adj):
    diff_mat = numpy.abs(diffs(nopis, nopis))
    mat = numpy.exp(-diff_mat)*(adj>0)
    # import matplotlib.pyplot as plt
    # plt.hist(mat[mat!=0],100)
    # # plt.plot(xxA, "b")
    # plt.show()

    L = -mat + numpy.diag(numpy.sum(mat,1))
    tord, tval = fiedler_ord(L)
    return tord

def fiedler_order(graph, adj):
    mat = adj # 1.0+0.1*(adj>0)
    L = -mat + numpy.diag(numpy.sum(mat,1))
    tord, tval = fiedler_ord(L)
    return tord

def fiedler_ord(L):
    # evalu, evect = scipy.sparse.linalg.eigsh(L, 4, which="SM")
    # evaluA, evectA = scipy.sparse.linalg.eigs(L, 4, which="SR", tol=10**-5)
    # evaluA, evectA = scipy.sparse.linalg.eigs(L, 4, which="SR")
    evalu, evect = numpy.linalg.eig(L)
    evect = numpy.real(evect)
    # evect = numpy.around(numpy.real(evect), 10)
    # evectA = numpy.real(evectA)
    # tmp = evect[:,-1]
    evalu[numpy.argmin(evalu)] = float("Inf")
    tmp = evect[:, numpy.argmin(evalu)]
    # tmp = numpy.absolute(evect[:,1])
    # tmp[numpy.where(tmp < 10**-7)[0]] = 0
    # pmin, pmax = numpy.argmin(tmp), numpy.argmax(tmp)
    # if pmin < pmax:
    #     ttt = numpy.argsort(tmp)
    # else:
    #     ttt = numpy.argsort(-tmp)
    ttt = numpy.argsort(tmp)
    return ttt, tmp


def fiedler_val(L, ks):
    evalu, evect = numpy.linalg.eig(L)
    evect = numpy.real(evect)
    evalu = numpy.real(evalu)
    ttt = numpy.argsort(evalu)
    return numpy.dot(evect[:, ttt[ks]], numpy.diag(evalu[ttt[ks]])) 



def generate_positions(graph, adj=None, nopis=None, locs=None, map_ntoi=None):
    N = len(graph)

    if locs is not None and map_ntoi is not None:
        nn = graph.keys()
        llat, llong = zip(*locs.values())
        boundaries = ((min(llat), max(llat)), (min(llong), max(llong)))
        pos = dict([(n, (0., 0.)) for i,n in enumerate(nn)] +
                   [(n+N, (0., 0.)) for i,n in enumerate(nn)])
        for n, loc in locs.items():
            pos[map_ntoi[n]] = ((loc[0]-boundaries[0][0])/(boundaries[0][1]-boundaries[0][0])*3 -1.5,
                                (loc[1]-boundaries[1][0])/(boundaries[1][1]-boundaries[1][0])*3 -1.5)
            pos[map_ntoi[n]+N] = pos[map_ntoi[n]]
        return pos

    if nopis is not None:
        # Qub = numpy.dot(numpy.linalg.inv(numpy.eye(adj.shape[0])-adj),dia)
        # nn = fiedler_order_opi(graph, nopis, adj)
        nn = fiedler_order_wei(graph, nopis, adj, {})
#        nn = fiedler_order_opi(graph, nopis, adj)
    elif adj is not None:
        nn = fiedler_order(graph, adj)
    else:
        nn = graph.keys()
        nn.sort(key=lambda x: len(graph[x]))
        # random.shuffle(nn)
    dw = 2*numpy.pi/N
    pos = dict([(n, (numpy.cos(dw*i),numpy.sin(dw*i))) for i,n in enumerate(nn)]+
               [(n+N, (1.05*numpy.cos(dw*i), 1.05*numpy.sin(dw*i))) for i,n in enumerate(nn)])
    return pos

def translate_data(graph):
    map_iton = sorted(graph.keys())
    map_ntoi = dict([(v,k) for (k,v) in enumerate(map_iton)])
    trans_graph = {}
    for n, neighs in graph.items():
        trans_graph[map_ntoi[n]] = {}
        for neigh, nns in neighs.items():
            trans_graph[map_ntoi[n]][map_ntoi[neigh]] = nns
    return trans_graph, map_iton, map_ntoi

def graph_to_mats(graph, nopis, comm, data_parameters):
    if nopis is not None:
        max_auth = nopis.shape[0]
    else:
        max_auth = float("Inf")
    if comm is None:
        ## comm =  sorted(graph.keys())
        comm =  sorted([c for c in graph.keys() if c < max_auth])
        map_nd = {}
    else:
        map_nd = dict([(v,k) for (k,v) in enumerate(comm)])
        
    selfw = 1
    N = len(comm)
    adj = numpy.zeros((N, N))
    dia = selfw*numpy.eye(N)

    for node in comm:
        for neigh, w in graph[node].items():
            ## if neigh in map_nd or len(map_nd) == 0:
            if neigh in map_nd or (len(map_nd) == 0 and neigh < max_auth):
                adj[map_nd.get(node, node), map_nd.get(neigh, neigh)] = 1.

    sumc = numpy.sum(adj,1)+numpy.sum(dia,1)
    adj = numpy.dot(numpy.eye(N)/sumc,adj)
    dia = numpy.dot(numpy.eye(N)/sumc,dia)
    return adj, dia, comm

def edges_to_mats(graph, nopis, edges, data_parameters):

    selfw = 1
    N = len(graph)
    adj = numpy.zeros((N, N))
    dia = selfw*numpy.eye(N)

    for edge in edges:
        try:
            adj[edge[0], edge[1]] = 1.
            adj[edge[1], edge[0]] = 1.
        except IndexError:
            pdb.set_trace()
            print edge
            
    sumc = numpy.sum(adj,1)+numpy.sum(dia,1)
    adj = numpy.dot(numpy.eye(N)/sumc,adj)
    dia = numpy.dot(numpy.eye(N)/sumc,dia)
    return adj, dia
    
def get_zzs_comm(adj, dia, ss, data_parameters):
    ## Qub = numpy.dot(numpy.linalg.inv(numpy.eye(adj.shape[0])-adj),dia)
    ## return numpy.dot(Qub, ss)
    # if numpy.sum(numpy.round(numpy.hstack((adj, dia)).sum(1), 5) != 1) > 0:
    #     sumc = numpy.sum(adj,1)+numpy.sum(dia,1)
    #     radj = numpy.dot(numpy.eye(adj.shape[0])/sumc,adj)
    #     rdia = numpy.dot(numpy.eye(dia.shape[0])/sumc,dia)
    # else:
    #     radj = adj
    #     rdia = dia

    if data_parameters.get("ownopiw", 0) > 0:
        ownopiw = float(data_parameters["ownopiw"])
        othopiw = 1-ownopiw
    else:
        ownopiw = 1.
        othopiw = 1.

    gmat = numpy.concatenate((othopiw*(adj>0), ownopiw*(dia>0))).T
    ssums = numpy.sum(gmat, axis=1)
    radj = numpy.divide(othopiw*(adj>0), numpy.tile(ssums, (ssums.shape[0], 1)).T)
    rdia = numpy.divide(ownopiw*(dia>0), numpy.tile(ssums, (ssums.shape[0], 1)).T)
    Qub = numpy.dot(numpy.linalg.inv(numpy.eye(radj.shape[0])-radj),rdia)
    zz = numpy.dot(Qub, ss)

    # zzi = get_zzs_comm_iterate(adj, dia, ss, data_parameters)
    # zzii = get_zzs_comm_iterate(adj, dia, ss, data_parameters, zz)
    # QubNW = numpy.dot(numpy.linalg.inv(numpy.eye(adj.shape[0])-adj),dia)
    # zzNW = numpy.dot(QubNW, ss)
    # print "----------------------\nunw=%s\norg=%s\nzit=%s\niit=%s" % (zzNW, zz, zzi, zzii)
    return zz

def get_zzs_comm_iterate(adj, dia, ss, data_parameters, ini_op=None):
    if ini_op is None:
        ini_op = ss
    if dia is None:
        dia = numpy.eye(adj.shape)
    if data_parameters.get("ownopiw", 0) > 0:
        ownopiw = float(data_parameters["ownopiw"])
        othopiw = 1-ownopiw
    else:
        ownopiw = 1.
        othopiw = 1.
    
    gmat = numpy.concatenate((othopiw*(adj>0), ownopiw*(dia>0))).T
    ssums = numpy.sum(gmat, axis=1)
    old_op = -1.*numpy.ones(ini_op.shape)
    it_count = 0
    while numpy.sum((ini_op-old_op)**2) > ZIT_EPS and it_count < ZIT_NB:
        it_count += 1
        # if it_count % 10 == 0:
        #     print "---\t", it_count, numpy.sum((ini_op-old_op)**2), numpy.min(ini_op), numpy.max(ini_op)
        old_op = ini_op
        ini_op = numpy.divide(numpy.dot(gmat, numpy.concatenate((old_op, ss))),ssums)
    #print "Recomputed opis", it_count
    # print "STOPPED  eps=%s (%s), itc=%d (%d)" % (numpy.sum((ini_op-old_op)**2), ZIT_EPS,  it_count, ZIT_NB)
    return ini_op

def max_edges(ss, adj):
    diff_mat = numpy.tile(ss, (ss.shape[0],1,1))
    diff_mat -= diff_mat.transpose((1,0,2))
    return numpy.max(numpy.array([(adj > 0)*diff_mat[:,:,i] for i in range(diff_mat.shape[2])]))
    diff_mat = (adj > 0)*(numpy.outer(numpy.ones(ss.shape[0]), ss) - numpy.outer(ss, numpy.ones(ss.shape[0])))
    return numpy.max(diff_mat)

def noavg_opd(ss, adj):
    inds = numpy.where(adj > 0)
    return (ss[inds[0]] - ss[inds[1]])**2 #).sum(1) ### or sum in stats?


# def get_stats_opd(opd):
#     return {"opd_max": numpy.max(opd),
#             "opd_mean": numpy.mean(opd),
#             "opd_median": numpy.median(opd),
#             "opd_var": numpy.var(opd),
#             "opd_tot": numpy.sum(opd),
#             "opd_len": opd.shape[0]}
    
def get_stats_opd(opdd):
    opd  = opdd.sum(1)
    return {"opd_max": numpy.max(opd),
            "opd_mean": numpy.mean(opd),
            "opd_median": numpy.median(opd),
            "opd_var": numpy.var(opd),
            "opd_tot": numpy.sum(opd),
            "opd_len": opdd.shape[0],
            "opd_depth": opdd.shape[1]}

def part_tension(ss, zz, adj, dia):
    diff_mat = diffs(ss, zz)        
    ## inn = numpy.diag(diff_mat)
    ## cross = (adj > 0)*diff_mat
    inn = numpy.array([numpy.diag(diff_mat[:,:,i]) for i in range(diff_mat.shape[2])])
    cross = numpy.array([(adj > 0)*diff_mat[:,:,i] for i in range(diff_mat.shape[2])])
    return numpy.sum(inn**2), numpy.sum(cross**2)

def all_tension(ss, zz, adj, dia, alpha=1.):
    Ui, Uc = part_tension(ss, zz, adj, dia)
    return alpha*Ui+(2-alpha)*Uc

def diffs(innate_beliefs, expressed_beliefs):
    # nb = len(expressed_beliefs)
    # diff_mat = numpy.outer(numpy.ones(nb), expressed_beliefs)
    # diff_mat[range(nb), range(nb)] = innate_beliefs
    # diff_mat -= numpy.outer(expressed_beliefs, numpy.ones(nb))
    # return diff_mat
    nb = expressed_beliefs.shape[0]
    diff_mat = numpy.tile(expressed_beliefs, (nb,1,1))
    diff_matB = diff_mat.copy().transpose((1,0,2))
    for i in range(diff_mat.shape[2]):
        diff_mat[range(nb), range(nb), i] = innate_beliefs[:, i]
    diff_mat -= diff_matB
    return diff_mat

def getOpiDiffGraph(graph, nopis):
    opidiff_graph = {} 
    for n, neighs in graph.items():
        opidiff_graph[n] = dict([(neigh, abs(nopis[n]-nopis[neigh])) for neigh in neighs.keys()])
    return opidiff_graph

def getOpiAvg(graph, nopis):
    popis = numpy.zeros_like(nopis)
    for n, nds in graph.items():
        popis[n] = numpy.mean(nopis[[n]+nds.keys()])
    return popis

def getOpiDiffEdgesAvg(graph, nopis, popis):
    opidiff_edges = SortedList() 
    for n, neighs in graph.items():
        opidiff_edges.update([((popis[n]-popis[neigh])**2+
                               (nopis[n]-popis[n])**2+
                               (nopis[neigh]-popis[neigh])**2, n, neigh) for neigh in neighs.keys() if n < neigh])
    return opidiff_edges

def getOpiDiffEdges(graph, nopis):
    opidiff_edges = SortedDict() 
    for n, neighs in graph.items():
        opidiff_edges.update([((1-abs(nopis[n]-nopis[neigh]), n, neigh), n) for neigh in neighs.keys() if n < neigh])
    return opidiff_edges

def getOpiDiffEdgesL(graph, nopis):
    opidiff_edges = []
    for n, neighs in graph.items():
        opidiff_edges.extend([(abs(nopis[n]-nopis[neigh]), n, neigh) for neigh in neighs.keys() if n < neigh])
    return opidiff_edges


def getOpiDiffNodes(graph, nopis):
    opidiff_nds = {}
    for n, neighs in graph.items():
        opidiff_nds[n] = numpy.sum([abs(nopis[n]-nopis[neigh]) for neigh in neighs.keys()])
    return opidiff_nds

def getKMinValI(candidates, graph, nopis, commc):
    peaks = findPeaks(graph, nopis, nodeset=set(commc), cands=candidates.keys())
    if len(peaks[0])+len(peaks[1]) == len(candidates):
        return getKMinVal(candidates)
    peaks = [peaks[0], peaks[1], set(candidates.keys()).difference(peaks[0]+peaks[1])]
    mins = [(float("Inf"), None), (float("Inf"), None), (float("Inf"), None)]
    for c in range(3):
        for k in peaks[c]:
            v = candidates[k]
            if v < mins[c][0] or (v == mins[c][0] and k < mins[c][1]):
                mins[c] = (v, k)
    # if len(peaks[-1]) > 0:
    #     pdb.set_trace()
    kkk = sorted([(k[0], k[1],i) for i,k in enumerate(mins)])[0][2]
    return mins[-1][0], mins[-1][1]
    return mins[kkk][0], mins[kkk][1]

def getKMinVal(candidates):
    minv, mink = (float("Inf"), None) 
    for k,v in candidates.iteritems():
        if v < minv or (v == minv and k < mink):
            minv, mink = v, k
    return minv, mink



def get_connected_k(graph, k):
    seed = numpy.random.choice(graph.keys())
    subg = set()
    border = set([seed])
    while len(subg)+len(border) < k:
        touch = set().union(*[graph[n].keys() for n in border])
        touch.difference_update(*[subg, border])
        subg.update(border)
        border = touch
    while len(subg) < k and len(border) > 0:
        subg.add(border.pop())
    return subg


def findPeaks(graph, nopis, nodeset=None, cands=None):
    tops = []
    lows = []
    if cands is None:
        if nodeset is None:
            ks = graph.keys()
        else:
            ks = nodeset
    else:
        ks = cands
    for nd in ks:
        if nodeset is None:
            neighs = graph[nd].keys()
        else:
            neighs = list(nodeset.intersection(graph[nd].keys()))
        if nopis[nd] > numpy.max(nopis[neighs]):
            tops.append(nd)
        elif nopis[nd] < numpy.min(nopis[neighs]):
            lows.append(nd)
    return tops, lows

def eval_edges(graph, nopis, edges, data_parameters):
    adj, dia = edges_to_mats(graph, nopis, edges, data_parameters)
    zz = get_zzs_comm(adj, dia, nopis, data_parameters)
    Ui, Uc = part_tension(nopis, zz, adj, dia)
    U = Ui+Uc
    return Ui, Uc, U

##############################################################
############# POSTPROCESS, RE-EVALUATION
def revalTeamsQ(graph, nopis, data_parameters):
    patt_summaries = re.sub("reval_.*_summary","trunq_*_summary", data_parameters["gasis"])

    all_opd = numpy.array([(nopis[nA]-nopis[nB])**2 for (nA, nB) in itertools.chain(*[[(n,ne) for ne in sorted(graph[n].keys()) if n < ne] for n in sorted(graph.keys())])])
    all_opd_stats = get_stats_opd(all_opd)
    for summary in glob.glob(patt_summaries):
        numb = re.search("trunq_(?P<numb>[a-zA-Z0-9_]*)_summary", summary).group("numb")
        foutname = re.sub("reval_.*_summary","reval_%s_summary" % numb , data_parameters["gasis"])
        with open(foutname, "w") as fo:
            with open(summary) as fp:
                head = None
                for line in fp:
                    if head is None:
                        head = dict([(v,k) for (k,v) in enumerate(line.strip().split())])
                        fo.write(line.strip() +
                                 "\tsol_opd_max\tsol_opd_mean\tsol_opd_median\tsol_opd_tot\tsol_opd_var\tsol_opd_len" +
                                 "\tall_opd_max\tall_opd_mean\tall_opd_median\tall_opd_tot\tall_opd_var\tall_opd_len\n")

                    else:
                        parts = line.strip().split()
                        teamfile = "%s%s.%s_%s.communities" % (data_parameters['results_rep'], parts[head["basis"]],
                                                          parts[head["external"]], parts[head["qstr"]])
                        details = loadTeamQ(teamfile)
                        if details is not None:
                            nodeset = set(details["num_nodes"])
                            edgeset = sorted(set().union(*[[(n,ne) for ne in nodeset.intersection(graph[n].keys()) if n < ne] for n in nodeset]))
                            sol_opd = numpy.array([(nopis[nA]-nopis[nB])**2 for (nA, nB) in edgeset])
                            sol_opd_stats = get_stats_opd(sol_opd)
                            fo.write(line.strip() +
                                     ("\t%(opd_max)f\t%(opd_mean)f\t%(opd_median)f\t%(opd_tot)f\t%(opd_var)f\t%(opd_len)f" % sol_opd_stats)+
                                     ("\t%(opd_max)f\t%(opd_mean)f\t%(opd_median)f\t%(opd_tot)f\t%(opd_var)f\t%(opd_len)f\n" % all_opd_stats))


def revalTeamsQOW(graph, nopis, data_parameters):
    patt_summaries = re.sub("reval_.*_summary","trunq_*_summary", data_parameters["gasis"])

    for summary in glob.glob(patt_summaries):
        numb = re.search("trunq_(?P<numb>[a-zA-Z0-9_]*)_summary", summary).group("numb")
        foutname = re.sub("reval_.*_summary","reval_%s_summary" % numb , data_parameters["gasis"])
        foutname = re.sub("xps_divers","ow_divers", foutname)
        with open(foutname, "w") as fo:
            with open(summary) as fp:
                head = None
                for line in fp:
                    if head is None:
                        head = dict([(v,k) for (k,v) in enumerate(line.strip().split())])
                        fo.write(line.strip() +
                                 "\townopiw\n")

                    else:
                        parts = line.strip().split()
                        teamfile = "%s%s.%s_%s.communities" % (data_parameters['results_rep'], parts[head["basis"]],
                                                          parts[head["external"]], parts[head["qstr"]])
                        details = loadTeamQ(teamfile)
                        if details is not None and "num_nodes" in details:
                            curr_idxs = details["num_nodes"]
                            (adjG, diaG, ccc) = data_parameters["mats"]
                            for ownopiw in data_parameters["ownopiws"]:
                                data_parameters["ownopiw"] = ownopiw
                                itzz = get_zzs_comm(adjG[:, curr_idxs][curr_idxs,:], diaG[:, curr_idxs][curr_idxs,:], details["org_opis"], data_parameters)
                                Ui, Uc = part_tension(details["org_opis"], itzz, adjG[:, curr_idxs][curr_idxs,:],
                                                      diaG[:, curr_idxs][curr_idxs,:])
                                
                                nparts = ["%f" % v for v in [Ui+Uc, Ui, Uc]]+ parts[3:]+["%f" % ownopiw]
                                fo.write("\t".join(nparts)+"\n")

            
def loadTeamQ(teamfile):
    with open(teamfile, "r") as fp:
        line = fp.readline()
        tmp0 = re.match("------------METHOD=(?P<meth>[a-zA-Z]*)-(?P<score>[a-zA-Z]) k=(?P<lq>[0-9]*):(?P<Q>[0-9\-]*)\n", line)
        line = fp.readline()
        tmp = re.match("U=(?P<U>[0-9.]*) Ui=(?P<Ui>[0-9.]*) Uc=(?P<Uc>[0-9.]*) N=(?P<N>[0-9]*) E=(?P<E>[0-9.]*)\n", line)
        if tmp0 is not None and tmp is not None:
            details = {"meth": tmp0.group("meth"), "score": tmp0.group("score"), "lq": tmp0.group("lq"),"Q": map(int,tmp0.group("Q").split("-")),  
                       "U": float(tmp.group("U")), "Ui": float(tmp.group("Ui")), "Uc": float(tmp.group("Uc")),
                       "N": int(tmp.group("N")), "E": int(tmp.group("E"))}
            line = fp.readline()
            # details["str_nodes"] = line.strip().split("\t")
            # line = fp.readline()
            ttn = re.match("^.* opd_len=[1-9][0-9]*.000000(?P<nodes>[0-9].*)$", line)
            if ttn is not None:
                details["num_nodes"] = map(int, ttn.group("nodes").strip().split("\t"))
            else:
                pdb.set_trace()
            line = fp.readline()
            details["org_opis"] = map(float, line.strip().split("\t"))
            line = fp.readline()
            details["fin_opis"] = map(float, line.strip().split("\t"))
            return details

