#!/usr/bin/env python
import itertools, copy
import numpy as np
import scipy.sparse
import networkx as nx
import random
import matplotlib.pyplot as plt
import pdb

# ttt = np.vstack([np.hstack([np.ones(N) -np.eye(N), np.concatenate([np.zeros((N-1,1)), [[1]]])]), np.concatenate([np.zeros((N-1,1)), [[1],[0]]]).T])
# ttt = np.dot(np.diag(1/(1+np.sum(ttt, axis =1))), ttt)


def diffs(innate_beliefs, expressed_beliefs):
    nb = len(expressed_beliefs)
    diff_mat = np.outer(np.ones(nb), expressed_beliefs)
    diff_mat[range(nb), range(nb)] = innate_beliefs
    diff_mat -= np.outer(expressed_beliefs, np.ones(nb))
    return diff_mat

def disp(adj, dia, ss):
    print "---------------"
    print "adj=", adj
    print "dia=", dia
    print "ss=", ss
    print "---------------"

def get_circ_pos(nn, N=None):
    if N is None:
        N = nn
    dw = 2*np.pi/N
    return [(np.cos(dw*i),np.sin(dw*i)) for i in range(nn)], [(1.3*np.cos(dw*i), 1.3*np.sin(dw*i)) for i in range(nn)]

def get_lin_pos(nn, N=None):
    if N is None:
        N=0
    return [(N+i,0) for i in range(nn)], [(N+i,0.3) for i in range(nn)]

def get_lin_mat(N):
    return np.diag(np.ones(N-1), 1) + np.diag(np.ones(N-1), -1)

def get_circ_mat(N):
    mat = get_lin_mat(N)
    mat[0,N-1] = 1
    mat[N-1,0] = 1
    return mat

def build_mats(ttt, ss):
    N = len(ss)
    adj = ttt
    adj[range(N), range(N)] = 0
    sumc = np.sum(adj,1) + 1 #np.diag(ttt)*absp
    adj = np.dot(np.eye(N)/sumc,adj)
    dia = np.eye(N)/sumc
    return adj, dia

def get_opinions_inv(adj, dia, ss):
    Qub = np.dot(np.linalg.inv(np.eye(len(ss))-adj),dia)
    zz = np.dot(Qub, ss)
    return zz

def get_opinions_soc(adj, dia, ss):
    badj = 1*(adj>0)
    Qub = np.linalg.inv(np.diag(2*np.sum(badj,1)+1)-2*badj)
    zz = np.dot(Qub, ss)
    return zz


def get_opinions(adj, dia, ss):
    zz = np.zeros_like(ss)
    nzz = ss
    degrees = np.sum(adj>0, 1)+1
    while np.sum((zz-nzz)**2) > 10**-5:
        zz = nzz
        nzz = (np.dot(adj>0, zz)+ss)/degrees
    return nzz

def eval_sample(adj, dia, ss, which="inv"):
    if which == "":
        zz = get_opinions(adj, dia, ss)
    elif which == "soc":
        zz = get_opinions_soc(adj, dia, ss)
    else:
        zz = get_opinions_inv(adj, dia, ss)
    diff_mat = diffs(ss, zz)
    tmpc = (adj > 0)*diff_mat
    tmpi = (dia > 0)*diff_mat
    ui = np.sum(tmpi**2)
    uc = np.sum(tmpc**2)
    u_all = ui+uc
    return zz, u_all, ui, uc


def complot_sample(ki, ttt, ss, pos, confid, axes, tops, bounds, which="inv"):
    N = len(ss)
    adj, dia = build_mats(ttt, ss)
    zz, u_all, ui, uc = eval_sample(adj, dia, ss, which)
    zzo, u_allo, uio, uco = eval_sample(adj, dia, ss, "soc")
    
    fad = np.vstack([np.hstack([adj, dia]), np.hstack([np.zeros((N,N)), np.eye(N)])])
    G = nx.DiGraph(fad)

    ctid = 0
    if confid.get("si",0) < len(tops):
        ctid = confid.get("si",0) 
    for code, sc in [("i", ui), ("c", uc), ("u", u_all)]:
        if code in tops[ctid]:
            if sc < tops[ctid][code][0]:
                tops[ctid][code] = (sc, axes)

    ### PRINT HERE     
    print "%f %f\tu=%f i=%f c=%f\tZZ=%f = %s" % (u_all/u_allo, 9/8., u_all, ui, uc, sum(zz), "0") #zz)
    # print "%f\t%f\t%f" % (u_all, ui, uc)

    nx.draw_networkx(G, node_color=np.concatenate([zz, ss]), pos=pos, node_size=150, with_labels=False,cmap=plt.cm.Blues, vmin=0, vmax=1, ax=axes)
    # C = sum([abs(ss[i]-ss[i+1])**2 for i in range(len(ss)-1)])
    # axes.set_xlabel("i=%f c=%f\n%f" % (ui, uc, u_all))
    axes.set_xlabel("i=%.4f c=%.4f\nu=%.4f" % (ui, uc, u_all))
    # axes.set_xlabel("%f" % u_all)
    axes.set_xticks([])
    axes.set_yticks([])
    axes.set_xlim(bounds[0])
    axes.set_ylim(bounds[1])
    # axes.set_title(confid.get("title", ""))
    # if confid.get("sj", 0) == 1: 
    #     # axes.set_title(confid.get("title", ""))
    #     axes.set_title(",".join(["%.1f" % v for v in ss]))
    if confid.get("sj", -1) == 0:
        lbb = ",".join([("%.2f" % v)+(ssi+1==len(ss)/2)*"\n" for ssi,v in enumerate(ss)])
        axes.set_ylabel(lbb)

def add_tops(tops):
    styles = {"c": ("right", "red"), "i": ("left", "blue"), "u": ("top", "#551A8B")}
    for top in tops:
        for code, (sc, axes) in top.items():
            if code in styles:
                axes.spines[styles[code][0]].set_color(styles[code][1])
                for spine in axes.spines.values():
                    spine.set_linewidth(2)
            

def random_exp():

    # #### BASIC CIRCLE EXAMPLES
    nn = 10

    # sss = [np.array([1*(i==nn) for i in range(nn+1)]),
    #        np.array([1 for i in range(nn+1)])]
    # sss = [[0,0,1,1,0.33,0.66, 0.5],
    #        [0,0,0.33,0.5,0.66,1,1],
    #        [0,0,0,0.5,1,1,1]]

    # sss = [[0.2,0.3,0.4,0.5,0.6,0.7,0.8],
    #        [0,0.1,0.5,0.5,0.5,0.9,1],
    #        [0,0,0,1,1,1,0.5]]
    # sss = [[0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0],
    #        [0,0.1,0.2,0.3,0.5,0.5,0.5,0.7,0.8,0.9,1],
    sss = [[0,0,0,0,0,1,1,1,1,1,0.5],
           [0,0,0,0,0,1,1,1,1,1,0.501],
           [0,0,0,0,0.2,0.8,1,1,1,1,0.5],
           [0,0,0,0,0.2,0.8,1,1,1,1,0.25]]

    sss = [[0,0,0,0,0,0,1,1,1,1,1,1,0.5],
           [0,0,0,0,0,0,1,1,1,1,1,1,0.51],
           [0,0,0,0,0,0.1,0.9,1,1,1,1,1,0.5],
           [0,0,0,0,0.1,0.1,0.9,0.9,1,1,1,1,0.25]]

    sss = [[0,1]*(nn/2),
           [0,0.01,0.02,0.03,0.04,.96,.97,.98,.99,1],
           [0 for i in range(nn/2)]+[1 for i in range(nn/2)],
           [0 for i in range(nn-1)]+[1],
           [0 for i in range(nn-2)]+[0.1, 1],
           [i/(nn-1.) for i in range(nn)]]
           # [0,0,0,0,0,0,1,1,1,1,1,1],
           # [0,0,0,0,0,0.1,0.9,1,1,1,1,1],

    sss = [[0,0.01,0.02,0.03,0.04,.96,.97,.98,.99,1], 
           [0 for i in range(nn-1)]+[1],
           [i/(nn-1.) for i in range(nn)],
           sorted([np.random.rand() for i in range(nn)])]

    sss = [[0,0.01,0.02,0.03,0.04,.96,.97,.98,.99,1], 
           [i/(nn-1.) for i in range(nn)],
           sorted([np.random.rand() for i in range(nn)])]


    # sss = [[0,0,0,0,0,0,0,1,1,1],
    #        [0,0.005,0.01,0.015,0.02,0.025,0.03,0.98,0.99,1],
    #        [i/(nn-1.) for i in range(nn)]]



    # sss = [[0,0.01,0.02,0.03,0.04,.96,.97,.98,.99,1],
    #        [0 for i in range(nn-1)]+[1],
    #        [i/(nn-1.) for i in range(nn)],
    #        sorted([np.random.rand() for i in range(nn)])]

           # [0,0,0,0,0,0,1,1,1,1,1,1],
           # [0,0,0,0,0,0.1,0.9,1,1,1,1,1],

    # sss = [[0,0,0,0,0,0,0,1,1,1,1,1,0.5],
    #        [0,0,0,0,0,1,1,1,1,1,0.5]]

    posO, posI = get_circ_pos(nn, nn)
    pos = dict(enumerate(posO+[(0,0.1)]+posI+[(0.3,0.1)]))
    pos = dict(enumerate(posO+posI))

    mats = {}
    # mats["line"] = get_lin_mat(nn+1)
    
    #### MAT STAR
    # mats["star"] = np.zeros((nn+1, nn+1))
    # mats["star"][:, nn] = 1
    # mats["star"][nn,:] = 1

    #### MAT LINE WITH CENTER IN THE MIDDLE 
    mats["lino"] = get_lin_mat(nn+1)
    mats["lino"] = get_lin_mat(nn)
    # mats["lino"][nn/2-1, nn/2] = 0
    # mats["lino"][nn/2, nn/2-1] = 0
    # mats["lino"][-2, -1] = 0
    # mats["lino"][-1, -2] = 0
    # mats["lino"][nn/2-1, -1] = 1
    # mats["lino"][-1, nn/2-1] = 1
    # mats["lino"][nn/2, -1] = 1
    # mats["lino"][-1, nn/2] = 1
 
    # mats["lino"][0,nn] += 1
    # mats["lino"][nn, 0] += 1

    # #### MAT CIRCLE OPEN WITH CENTER 
    # mats["lino"] = np.zeros((nn+1, nn+1))
    # mats["lino"][range(1,nn),range(nn-1)] += 1
    # mats["lino"][range(nn-1),range(1,nn)] += 1
    # mats["lino"][0,nn] += 1
    # mats["lino"][nn, 0] += 1
    
    # #### MAT CIRCLE WITH CENTER ODD
    # mats["cc"] = np.zeros((nn+1, nn+1))
    # mats["cc"][range(1,nn),range(nn-1)] += 1
    # mats["cc"][range(nn-1),range(1,nn)] += 1
    # mats["cc"][nn,1::2] += 1
    # mats["cc"][1::2,nn] += 1
    # mats["cc"][0,nn-1] += 1
    # mats["cc"][nn-1, 0] += 1

    #### MAT CLIQUE
    mats["acliq"] = np.ones((nn+1, nn+1))
    mats["acliq"] = np.ones((nn, nn))

    #### MAT CLIQUE plus 1
    # mats["clip"] = np.ones((nn+1, nn+1))
    # mats["clip"][:, nn] = 0
    # mats["clip"][0, nn] = 1
    # mats["clip"][nn,:] = 0
    # mats["clip"][nn,0] = 1

    mats["blop"] = np.ones((nn, nn))
    mats["blop"][:, nn-1] = 0
    mats["blop"][nn-2, nn-1] = 1
    mats["blop"][nn-1,:] = 0
    mats["blop"][nn-1,nn-2] = 1

    #### MAT TWO CLIQUES plus bridge
    pivot = 5
    mats["clip"] = np.zeros((nn, nn))
    for i in range(pivot):
        mats["clip"][range(pivot), i] = 1
    for i in range(pivot, nn):
        mats["clip"][range(pivot, nn), i] = 1
    mats["clip"][pivot-1, pivot] = 1
    mats["clip"][pivot, pivot-1] = 1

    # mats["clio"] = np.zeros((nn, nn)) + mats["clip"]
    # mats["clio"][0, -1] = 1
    # mats["clio"][-1, 0] = 1
    # # mats["clip"][nn/2, -1] = 1
    # # mats["clip"][-1, nn/2] = 1

    # mats["clia"] = np.zeros((nn, nn)) + mats["clip"]
    # mats["clia"][pivot-1, pivot+1] = 1
    # mats["clia"][pivot+1, pivot-1] = 1

    #### MAT CLIQUE plus 1
    # mats["claq"] = np.zeros((nn+1, nn+1))
    # mats["claq"] = np.zeros((nn, nn))
    # for i in range(nn/4,nn/2):
    #     mats["claq"][range(nn/2-1), i] = 1
    #     mats["claq"][i, range(nn/2-1)] = 1
    #     mats["claq"][[nn-j-1 for j in range(nn/2-1)], nn-i-1] = 1
    #     mats["claq"][nn-i-1, [nn-j-1 for j in range(nn/2-1)]] = 1
    # mats["claq"][nn/2-1, nn/2-2] = 1
    # mats["claq"][nn/2-2, nn/2-1] = 1
    # mats["claq"][nn/2, nn/2+1] = 1
    # mats["claq"][nn/2+1, nn/2] = 1
    # mats["claq"][nn/2, nn/2-1] = 1
    # mats["claq"][nn/2-1, nn/2] = 1
    # mats["claq"][nn/2-1, -1] = 1
    # mats["claq"][-1, nn/2-1] = 1
    # mats["claq"][nn/2, -1] = 1
    # mats["claq"][-1, nn/2] = 1

    # #### MAT CLIQUE plus 1
    # mats["clip"] = np.zeros((nn+1, nn+1))
    # mats["clip"] = np.zeros((nn, nn))
    # for i in range(nn/2):
    #     mats["clip"][range(nn/2), i] = 1
    #     mats["clip"][[nn-j-1 for j in range(nn/2)], nn-i-1] = 1
    # mats["clip"][nn/2, nn/2-1] = 1
    # mats["clip"][nn/2-1, nn/2] = 1


    samples = []
    mtks = sorted(mats.keys())
    ctops = []
    for si, ssa in enumerate(sss):  
        ctops.append({"i": (float("Inf"), None), "c": (float("Inf"), None), "u": (float("Inf"), None)})
        for sj, mtk in enumerate(mtks):
            samples.append((mats[mtk], ssa, pos, {"title": mtk, "si": si, "sj": sj}))

    bounds = [[-1.5, 1.5], [-1.5, 1.5]]

    tops = copy.deepcopy(ctops)
    f, axarr = plt.subplots(len(sss), len(mtks))
    for ki, (ttt, ss, pos, confid) in enumerate(samples):
        complot_sample(ki, ttt, ss, pos, confid, axarr[confid["si"], confid["sj"]], tops, bounds)
    add_tops(tops)
    plt.suptitle("INV")

    # tops = copy.deepcopy(ctops)
    # f, axarr = plt.subplots(len(sss), len(mtks))
    # for ki, (ttt, ss, pos, confid) in enumerate(samples):
    #     complot_sample(ki, ttt, ss, pos, confid, axarr[confid["si"], confid["sj"]], tops, bounds, which="soc")
    # add_tops(tops)
    # plt.suptitle("SOC")
    #plt.tight_layout()
    plt.show()


def tri_exp():

    nn = 4
    
    mats = {}
    #### MAT STAR
    mats["BC"] = np.zeros((nn+1, nn+1))
    mats["BC"][:, nn] = 1
    mats["BC"][nn,:] = 1

    mats["AB"] = np.zeros((nn+1, nn+1))
    mats["AB"][:, nn] = 1
    mats["AB"][nn,:] = 1
    mats["AB"][0,nn] = 0
    mats["AB"][nn,0] = 0
    mats["AB"][0,1] = 1
    mats["AB"][1,0] = 1

    mats["AC"] = np.zeros((nn+1, nn+1))
    mats["AC"][:, nn] = 1
    mats["AC"][nn,:] = 1
    mats["AC"][1,nn] = 0
    mats["AC"][nn,1] = 0
    mats["AC"][0,1] = 1
    mats["AC"][1,0] = 1

    mats["tri"] = np.zeros((nn+1, nn+1))
    mats["tri"][:, nn] = 1
    mats["tri"][nn,:] = 1
    mats["tri"][0,1] = 1
    mats["tri"][1,0] = 1

    posO, posI = get_circ_pos(nn, nn)
    pos = dict(enumerate(posO+[(0,0)]+posI+[(-0.3,0.1)]))

    sss = [np.array([1*(i==nn) for i in range(nn+1)]), np.array([0.5*(i==nn) for i in range(nn+1)])]
    # [0.34, 0.36, 0.34, 0.34, 0.4, 0.4, 0.3],
    # [0.34, 0.36, 0.4, 0.4, 0.4, 0.4, 0.35],
    # [0, 1, 0.5, 0.5, 0.5, 0.5, 0.5],
    # [1, 1, 0., 0., 0., 0., 0.5],           
    # sss = [[0.1, 0.2, 0., 0., 0., 0., 0.],
    #        [0.1, 0.2, 0., 0., 0., 0., 0.1],
    #        [0.1, 0.2, 0., 0., 0., 0., 0.15],
    #        [0.1, 0.2, 0., 0., 0., 0., 0.2],
    #        [0.1, 0.2, 0., 0., 0., 0., 0.25]]

    sss = [[0.1, 0.2, 0.5, 0.5, 0., 0., 0.],
           [0.1, 0.2, 0.5, 0.5, 0., 0., 0.1],
           [0.1, 0.2, 0.5, 0.5, 0., 0., 0.15],
           [0.1, 0.2, 0.5, 0.5, 0., 0., 0.2],
           [0.1, 0.2, 0.5, 0.5, 0., 0., 0.25]]

    store = {}
    pcod = [2**i for i in range(len(mats))]
    pcod[2] = 2
    # for rndi in range(1000):
    # pdb.set_trace()
    # for kkk in itertools.product([0,1/3.,2/3.,1], repeat=nn+1):
    for kkk in itertools.product([0,0.2,0.4,0.6,0.8,1], repeat=nn+1):
        ss = np.array(kkk)
        # ss = np.random.rand(nn+1)
        tops = {"i": [float("Inf"), 0], "c": [float("Inf"), 0], "u": [float("Inf"), None]}
        for sj, (mt, mat) in enumerate(mats.items()):
            adj, dia = build_mats(mat, ss)
            zz, u_all, ui, uc = eval_sample(adj, dia, ss)
            # if np.sum((kkk - np.array([2/3., 2/3., 1/3., 1/3., 1]))**2) < 0.001:
            #     print kkk
            #     print ">>>", mt, u_all, ui, uc
            #     pdb.set_trace()
            for code, sc in [("i", ui), ("c", uc), ("u", u_all)]:
                if code in tops:
                    if (sc-tops[code][0])**2 < 10**-6:
                        tops[code][1] += pcod[sj]
                    elif sc < tops[code][0]:
                        tops[code] = [sc, pcod[sj]]
        # kkey = "-".join(["%d"% tops[cd][1] for cd in "icu"])
        if tops["i"][1] % 2 == 1:
            kkey = "-".join(["%d"% tops[cd][1] for cd in "icu"])
            if kkey not in store:
                store[kkey] = ss

    samples = []
    sss = store.values()
    # for si, ssa in enumerate(sss):
    for si, (kk, ssa) in enumerate(store.items()):  
        for sj, (mt, mat) in enumerate(mats.items()):
            samples.append((mat, ssa, pos, {"title": mt+kk, "si": si, "sj": sj}))
    

    bounds = [[-1.5, 1.5], [-1.5, 1.5]]
    tops = []
    for si in range(len(sss)):
        tops.append({"i": (float("Inf"), None), "c": (float("Inf"), None), "u": (float("Inf"), None)})
    f, axarr = plt.subplots(len(sss), len(mats))
    for ki, (ttt, ss, pos, confid) in enumerate(samples):
        complot_sample(ki, ttt, ss, pos, confid, axarr[confid["si"], confid["sj"]], tops, bounds)

    add_tops(tops)
    #plt.tight_layout()
    plt.show()

def ring_expA():

    samples = []
    for si, nn in enumerate([6,7]):

        mat_circle = get_circ_mat(nn)
        mat_circle[0,nn-1] = 0
        mat_circle[nn-1,0] = 0

        posO, posI = get_circ_pos(nn, nn)
        pos = dict(enumerate(posO+posI))

        ssa = np.array([1*(i==nn) for i in range(nn)])
        ssa = [r/float(nn-1) for r in range(nn)]
        ssb = ((nn+1)/2*ssa[:2])[:nn]

        samples.extend([(mat_circle, ssa, pos, {"si": si, "sj": 0}), (mat_circle, ssb, pos, {"si": si, "sj": 1})])

    bounds = [[-1.5, 1.5], [-1.5, 1.5]]
    f, axarr = plt.subplots(1, len(samples))
    tops = [{"i": (float("Inf"), None), "c": (float("Inf"), None), "u": (float("Inf"), None)}]
    for ki, (ttt, ss, pos, confid) in enumerate(samples):
        complot_sample(ki, ttt, ss, pos, confid, axarr[ki], tops, bounds)

    add_tops(tops)
    #plt.tight_layout()
    plt.show()

def ring_exp():

    samples = []
    nns = [6, 7]
    ssal = [r/float(max(nns)-1) for r in range(max(nns))]
    ssbl = ((max(nns)+1)/2*ssal[:2])

    for si, nn in enumerate(nns):

        mat_circle = get_circ_mat(nn)
        mat_circle[0,nn-1] = 0
        mat_circle[nn-1,0] = 0

        posO, posI = get_circ_pos(nn, nn)
        pos = dict(enumerate(posO+posI))

        # ssa = np.array([1*(i==nn) for i in range(nn)])
        # ssa = [r/float(nn-1) for r in range(nn)]
        # ssb = ((nn+1)/2*ssa[:2])[:nn]

        samples.extend([(mat_circle, ssal[:nn], pos, {"si": si, "sj": 0}), (mat_circle, ssbl[:nn], pos, {"si": si, "sj": 1})])

    bounds = [[-1.5, 1.5], [-1.5, 1.5]]
    f, axarr = plt.subplots(1, len(samples))
    tops = [{"i": (float("Inf"), None), "c": (float("Inf"), None), "u": (float("Inf"), None)}]
    for ki, (ttt, ss, pos, confid) in enumerate(samples):
        complot_sample(ki, ttt, ss, pos, confid, axarr[ki], tops, bounds)

    add_tops(tops)
    #plt.tight_layout()
    plt.show()

def line_exp():
    N = 5
    mat_line = get_lin_mat(N)
    posO, posI = get_lin_pos(N)
    print posO
    pos = dict(enumerate(posO+posI))
    
    # sss_mains = [[0,0.2,0,0.2,0,0.2], [0,0.2,.4,0.6,.8, 1], [0, 0.2, 0, 0.2, 0.4,.6]]
    sss_mains = [[0,0.2,0,0.2,0], [0,0.2,.4,0.6,.8], [0, 0.2, 0, 0.2, 0.4]]

    # subsets = [range(N), range(N)+[N+2]]
    samples = []
    for si, ssa in enumerate(sss_mains):
        samples.append((mat_line, ssa, pos,{"ss": ssa}))

    bounds = [[-1, N], [-.3, .6]]
    # plt.figure(figsize=(20,20))
    f, axarr = plt.subplots(1, len(samples))
    tops = [{"i": (float("Inf"), None), "c": (float("Inf"), None), "u": (float("Inf"), None)}]
    for ki, (ttt, ss, pos, confid) in enumerate(samples):
        complot_sample(ki, ttt, ss, pos, confid, axarr[ki], tops, bounds)

    add_tops(tops)
    #plt.tight_layout()
    plt.show()


def supp_line_exp():
    samples = []
    N = 5
    
    mat_line = np.diag(np.ones(N-1), 1) + np.diag(np.ones(N-1), -1)
    mat_linea = np.vstack([np.hstack([mat_line, np.diag(np.ones(N))]), np.hstack([np.diag(np.ones(N)), np.zeros((N, N))])]) 

    poso = [(i, 0) for i in range(N)] + [(i, -1) for i in range(N)]
    posi = [(i, 0.3) for i in range(N)] + [(i, -1.3) for i in range(N)]

    ss_main = [0,0,1,0,0]
    # sss_support = [[0,0,1,0,0], [0,0.1,0.7,0.1,0], [0.09,0.18,0.45,0.18,0.09]]
    # sss_support.append([(sss_support[0][jj]+sss_support[-1][jj])/2 for jj in range(len(sss_support[-1]))])
    # sss_support = [[0,0,1,0,0], [0.045, 0.09, 0.725, 0.09, 0.045],
    #                [0.0,0.1,0.8,0.1,0.0], [0.09,0.18,0.45,0.18,0.09],
    #                [0.09,0.18,0.40,0.18,0.09], [0.09,0.18,0.1,0.18,0.09], [0.0,0.05,0.1,0.05,0.0]]
    sss_support = [[0.0,0.1,0.8,0.1,0.0], 
                   [0.09,0.18,0.40,0.18,0.09], [0.09,0.18,0.1,0.18,0.09]]

    # ss_main = [0,1,1,1,0]
    # sss_support = [[0,1,1,1,0], [0.36, 0.72, 0.81, 0.72, 0.36]]
    # # sss_support = [[0,0,1,0,0], [0,0.1,0.7,0.1,0], [0.09,0.18,0.45,0.18,0.09]]
    # sss_support.append([(sss_support[0][jj]+sss_support[-1][jj])/2 for jj in range(len(sss_support[-1]))])
    # # sss_support = [[0,0,1,0,0], [0.045, 0.09, 0.725, 0.09, 0.045],
    # #                [0.0,0.1,0.7,0.1,0.0], [0.09,0.18,0.45,0.18,0.09]]

    subsets = [range(N), range(N)+[N+2], range(N)+[N, N+2, N+N-1], range(2*N)]
    # subsets = [range(N), range(N)+[N+2]]
    for si, ss_support in enumerate(sss_support):
        for sj, subset in enumerate(subsets):
            ssa = ss_main+ss_support
            mat_tmp = mat_linea[subset,:]
            samples.append((mat_tmp[:, subset], [ssa[s] for s in subset],
                            dict(enumerate([poso[s] for s in subset]+[posi[s] for s in subset])), {"si": si, "sj": sj}))      
 

#    for ssa in [(0,1,0), (1,0,1)]:
    # for ssa in [(0,0.49,0), (0,0.49,1), (0,0.51,0), (0,0.51,1)]:
    #     # ssa = [0,1,0] #np.random.rand(N)
    #     T = (1/32.+1/64.)*((3*ssa[0]-2*ssa[1]-ssa[2])**2+(3*ssa[2]-2*ssa[1]-ssa[0])**2) + (1/16.)*(2*ssa[1]-ssa[0]-ssa[2])**2
    #     C = sum([abs(ssa[i]-ssa[i+1])**2 for i in range(len(ssa)-1)])
    #     print T, C 
    #     samples.append((mat_line, ssa, pos, None))

    bounds = [[-1, N], [-1.6, .6]]
    # plt.figure(figsize=(20,20))
    f, axarr = plt.subplots(len(sss_support), len(subsets))
    tops = [{"i": (float("Inf"), None), "c": (float("Inf"), None), "u": (float("Inf"), None)}]
    for ki, (ttt, ss, pos, confid) in enumerate(samples):
        complot_sample(ki, ttt, ss, pos, confid, axarr[confid["si"], confid["sj"]], tops, bounds)

    add_tops(tops)
    #plt.tight_layout()
    plt.show()


def random_exp_alt():
    K = 4
    nb_nodes = 5
    nb_edges = None #15
    edge_p = 0.05
    absp = 0.3
    epsilon = 0.25
    gamma = 0.25

    # ttt = np.random.rand(N, N)
    # adj = (ttt < edge_p)*ttt
    # adj -= np.eye(N)*np.diag(adj)
    # sumc = np.sum(adj,1)
    # zz = np.where(sumc == 0)[0]
    # sumc[zz] = 1
    # adj[zz, zz] = 1.0
    # adj = np.dot(np.eye(N)/sumc,adj)
    # fad = adj

    # #### BASIC CIRCLE EXAMPLES
    # nn = 6
    # mat_circle = np.ones((nn+1, nn+1))
    # mat_circle[range(1,nn),range(nn-1)] -= 1
    # mat_circle[range(nn-1),range(1,nn)] -= 1
    # mat_circle[nn,1::2] -= 1
    # mat_circle[1::2,nn] -= 1
    # mat_circle[0,nn-1] -= 1
    # mat_circle[nn-1, 0] -= 1

    # ssa = np.array((nn/2+1)*[0, 1])

    # dw = 2*np.pi/nn
    # pos = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(N)])
    # pos = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(nn)]+
    #            [(i+nn+1, (1.3*np.cos(dw*i), 1.3*np.sin(dw*i))) for i in range(nn)])
    # pos[nn] = (0,0)
    # pos[2*nn+1] = (0.3,0)
    # pos[nn] = (0,0.1)
    # pos[2*nn+1] = (0.3,0.1)

    # # samples = [(mat_circle, ssa[:-1], pos), (mat_circle, ssa[1:], pos),
    # #            (mat_circle, ssa[range(nn)+[nn+1]], pos), (mat_circle, ssa[range(1,nn+1)+[nn]], pos),

    # samples = [(mat_circle, ssa[:-1], pos), (mat_circle, ssa[range(1,nn+1)+[nn]], pos),
    #            (mat_circle, np.concatenate([ssa[:nn],[0.5]]), pos),
    #            (mat_circle[:-1,:-1], ssa[:nn], None)]

    # samples = [(np.zeros((nn+1, nn+1)), np.concatenate([ssa[:nn],[0.5]]), pos),
    #            (np.zeros((nn+1, nn+1)), ssa[:nn+1], pos),
    #            (np.zeros((nn, nn)), ssa[:nn], None),
    #            ]

    #### positions for dropping nodes
    # dw = 2*np.pi/nn
    # pos1 = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(nn)]+
    #            [(i+nn, (1.3*np.cos(dw*i), 1.3*np.sin(dw*i))) for i in range(nn)])
    # pos2 = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(nn-1)]+
    #            [(i+nn-1, (1.3*np.cos(dw*i), 1.3*np.sin(dw*i))) for i in range(nn-1)])
    # pos3 = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(nn-2)]+
    #            [(i+nn-2, (1.3*np.cos(dw*i), 1.3*np.sin(dw*i))) for i in range(nn-2)])


    # ### FIRST REDUCTION COUNTER EXAMPLE 
    # O, N, M = (20, 5, 2)

    # # hitting_mats = [np.ones((N, M))]
    # # hitting_mats = [np.array([[0,0,1,1,1], [0,0,1,1,1], [0,1,0,1,1]])]
    # hitting_mats = [np.array([[0,0,0,1,1], [0,0,0,1,1]]),
    #                 np.array([[0,0,0,1,1], [1,1,0,0,0]])]
    # pos = dict([(i, (0,0)) for i in range(O)]+
    #            [(O+i, (1,i-N/2.0)) for i in range(N)]+
    #            [(N+O+i, (2,i-M/2.0)) for i in range(M)]+
    #            [(N+O+M+i, (-.25,0)) for i in range(O)]+
    #            [(N+O+M+O+i, (0.75,i-N/2.0)) for i in range(N)]+
    #            [(N+O+M+N+O+i, (2.25,i-M/2.0)) for i in range(M)])

    # samples = []
    # for hitting_mat in hitting_mats:
    #     mat = np.vstack([np.hstack([np.zeros((O, O)), np.zeros((O, N)), np.ones((O, M))]),
    #                      np.hstack([np.zeros((N, O)), np.ones((N, N)), hitting_mat.T]),
    #                      np.hstack([np.ones((M, O)), hitting_mat, np.ones((M, M))])])
    #     ss = np.concatenate([np.zeros(O+N), np.ones(M)])
    #     samples.append((mat, ss, pos))
    # samples = [(np.hstack([np.zeros((O, O)), np.zeros((O, N)), np.zeros((O, N))]), [1,0.5,0.5,0,0.5,0.5], pos1),
    #            (np.zeros((nn-1, nn-1)), np.concatenate([0.5*np.ones(nn-2),[1]]), pos2),
    #            (np.zeros((nn-2, nn-2)), 0.5*np.ones(nn-2), pos3),
    #            ]

    # #### RANDOM EXAMPLE
    # samples = [(np.random.rand(nb_nodes, nb_nodes), np.random.random(nb_nodes), None) for i in range(K)]

    ### SECOND REDUCTION TOY EXAMPLE
    ### [THE IMPORTANT MATRIX: COVER]
    hitting_mats = []
    hits = [[[0,1,2],[3,4,5],[6,7,8]], [[0,1,2],[0,3,4],[0,4,5],[6,7,8]]]
            # [[0,1,2],[0,1,2],[3,4,5],[6,7,8]],
            # [[0,1,2],[1,3,4],[0,4,5],[6,7,8]]]

    # hits = [[[0,1,2],[0,3,4],[0,4,5],[6,7,8]],
    #         [[0,1,2],[0,1,2],[3,4,5],[6,7,8]],
    #         [[0,1,2],[1,3,4],[0,4,5],[6,7,8]]]
    for hit in hits:
        N = max([max(h)+1 for h in hit])
        tmp = np.zeros((len(hit), N))
        for j, ids in enumerate(hit):
            tmp[j,ids] =1
        hitting_mats.append(tmp)

    samples = []
    (S, bL) = (N**2, N**2)
    # bsis = [S for nn in range(N)]


    nbd = 1
    samples = []
    for hi, hitting_mat in enumerate(hitting_mats):
        (M, N) = hitting_mat.shape
        for drop in range(-1, N):
            sis = [S for nn in range(N)]
            L = bL
            if drop == -1:
                L -= nbd
            else:
                sis[drop] -= nbd
            
            Ts = np.sum(sis)
            ss = np.concatenate([np.zeros(N), np.ones(M), np.zeros(Ts), np.ones(L)])
            mat = np.vstack([np.hstack([np.zeros((N, N)), hitting_mat.T, np.zeros((N, Ts)), np.zeros((N, L))]),
                             np.hstack([hitting_mat, np.ones((M, M)), np.zeros((M, Ts)), np.ones((M, L))])]+
                            [np.hstack([np.zeros((sis[si], N+M+np.sum(sis[:si]))), np.ones((sis[si], sis[si])), np.zeros((sis[si], np.sum(sis[si+1:]))), np.zeros((sis[si], L))]) for si in range(len(sis))]+
                            [np.hstack([np.zeros((L, N)), np.ones((L, M)), np.zeros((L, Ts)), np.ones((L, L))])]
                            )
            for si in range(len(sis)):
                mat[si, N+M+sum(sis[:si]):N+M+sum(sis[:si+1])] = 1
                mat[N+M+sum(sis[:si]):N+M+sum(sis[:si+1]), si] = 1    

            posl = [(i, (1,i-N/2.0)) for i in range(N)] + \
                   [(N+i, (-1,i-M/2.0)) for i in range(M)]
            for si in range(N):
                posl.extend([(N+M+sum(sis[:si])+i, (2,si-N/2.0)) for i in range(sis[si])])
            posl.extend([(N+M+sum(sis)+i, (-2,0)) for i in range(L)])
            pos = dict(posl+[(po[0]+len(posl), (po[1][0]+np.sign(po[1][0])*0.3, po[1][1]+0.3)) for po in posl])

            samples.append((mat, ss, pos, (N, M, S, bL, hi, drop)))

    plt.figure()
    # plt.figure(figsize=(9,3))
    for ki, (ttt, ss, pos, confid) in enumerate(samples):
        N = len(ss)
        # adj = (ttt < edge_p)*ttt
        # adj -= np.eye(N)*np.diag(adj)
        # ttt[range(N), range(N)] = 1.1
        # if nb_edges is not None:
        #     edge_p = np.sort(ttt, axis=None)[nb_edges]
        # adj = (ttt < edge_p)*np.ones((N,N))
        adj = ttt
        adj[range(N), range(N)] = 0
        sumc = np.sum(adj,1) + 1 #np.diag(ttt)*absp
        adj = np.dot(np.eye(N)/sumc,adj)
        dia = np.eye(N)/sumc

        fad = np.vstack([np.hstack([adj, dia]), np.hstack([np.zeros((N,N)), np.eye(N)])])
        G = nx.DiGraph(fad)
        if pos is None:
            dw = 2*np.pi/N
            # pos = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(N)])
            pos = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(N)]+
                       [(i+N, (1.3*np.cos(dw*i), 1.3*np.sin(dw*i))) for i in range(N)])

        Qub = np.dot(np.linalg.inv(np.eye(N)-adj),dia)
        zz = np.dot(Qub, ss)

        diff_mat = diffs(ss, zz)
        tmp = ((adj + dia) > 0)*diff_mat
        u_all = np.sum(tmp**2)

        # sumco = np.sum(adj[:-1,:-1]+dia[:-1,:-1], axis=1)
        # adjo = np.dot(np.diag(1/sumco),adj[:-1,:-1])
        # diao = np.dot(np.diag(1/sumco),dia[:-1,:-1])
        # Qubo = np.dot(np.linalg.inv(np.eye(N-1)-adjo),diao)
        # zzo = np.dot(Qubo, ss[:-1])

        # diff_mato = diffs(ss[:-1], zzo)
        # tmpo = ((adj[:-1,:-1] + dia[:-1,:-1]) > 0)*diff_mato
        # u_allo = np.sum(tmpo**2)

        # print drop, u_all, zz[:confid[0]+confid[1]], len(set(np.round(zz[:NNN+M], 4))), len(set(np.round(zz, 4)))
        susi = -1
        if confid[-1] >= 0:
            susi =np.sum(adj[confid[-1],:confid[0]+confid[1]] > 0)
        print confid[-2], confid[-1], u_all, len(set(np.round(zz[:confid[0]+confid[1]], 5))), len(set(np.round(zz, 5))), susi
        # # print u_allo
        # print zz
        # print zzo

        if confid[-1] == -1:
            plt.subplot(1, len(hitting_mats), confid[-2]+1)
            # plt.figure()
            nx.draw_networkx(G, node_color=np.concatenate([zz, ss]), pos=pos, node_size=150, with_labels=False,cmap=plt.cm.Blues, vmin=0, vmax=1)
            plt.xlabel("U = %f" % u_all)
            plt.xticks([])
            plt.yticks([])
        
        # plt.subplot(2, len(samples), len(samples)+ ki+1)
        # fado = np.vstack([np.hstack([adjo, diao]), np.hstack([np.zeros((N-1,N-1)), np.eye(N-1)])])
        # H = nx.DiGraph(fado)
        # dw = 2*np.pi/(N-1)
        # poso = dict([(i, (np.cos(dw*i),np.sin(dw*i))) for i in range(N-1)]+
        #             [(i+N-1, (1.3*np.cos(dw*i), 1.3*np.sin(dw*i))) for i in range(N-1)])
        # nx.draw_networkx(H, node_color=np.concatenate([zzo, ss[:-1]]), pos=poso, node_size=50, with_labels=False,cmap=plt.cm.Blues)
        # plt.xlabel("U = %f" % u_allo)
        # plt.xticks([])
        # plt.yticks([])

    plt.show()


def clique_bestdrop():
    N = 5
    edge_p = 0.35

    # ss = np.array([ 0.98334578,  0.49794024,  0.47450198,  0.91359232,  0.42755518]) #,
    # 0.55998235,  0.1938721 ,  0.52939444,  0.04228266,  0.24199978])
    ss = np.random.random(N)

    adj = (np.ones((N,N)) - np.eye(N))/(1.0*N)
    dia = np.eye(N)/(1.0*N)

    Qub = np.dot(np.linalg.inv(np.eye(N)-adj),dia)
    zz = np.dot(Qub, ss)

    diff_mat = diffs(ss, zz)
    u_all = np.sum(diff_mat**2)

    adjp = (np.ones((N-1,N-1)) - np.eye(N-1))/(1.0*(N-1))
    diap = np.eye(N-1)/(1.0*(N-1))

    Qubp = np.dot(np.linalg.inv(np.eye(N-1)-adjp),diap)

    u_alls = []
    u_alls_alt = []
    for out in range(len(ss)):
        nodes = range(N)
        ss0 = ss[[n for n in nodes if n != out]]
        zz0 = np.dot(Qubp, ss0)
        diff_mat0 = diffs(ss0, zz0)
        u_alls.append(np.sum(diff_mat0**2))
        u_alls_alt.append(((N+1.0)/N**2)*((N-1)*np.sum(ss0**2)-np.sum(np.outer(ss0, ss0))))
    
    min_alt = np.argmin(u_alls_alt)
    min_a = np.argmin(u_alls)
    min_s = np.argmax(N*ss**2-2*ss*np.sum(ss))
    if min_alt != min_a or min_alt != min_s:
        print "Oups!"
        pdb.set_trace()
    else:
        print "ok"

def clique_greedy():
    N = 8

    # ss = np.array([ 0.98334578,  0.49794024,  0.47450198,  0.91359232,  0.42755518]) #,
    ss = np.array([0.51, 0.52, 0.53, 0.6, 0.27, 0.255, 0.26]) #,
    N = len(ss)
    # 0.55998235,  0.1938721 ,  0.52939444,  0.04228266,  0.24199978])
    #ss = np.random.random(N)

    u_alls_seq = []
    rem_seq = []
    idx = range(len(ss))
    while len(idx) > 0:
        u_alls_seq.append((((len(idx)+2.0)/(len(idx)+1)**2)*(len(idx)*np.sum(ss[idx]**2)-np.sum(np.outer(ss[idx], ss[idx])))))
        min_s = np.argmax(len(idx)*ss[idx]**2-2*ss[idx]*np.sum(ss[idx]))
        rem_seq.append(idx.pop(min_s))

    u_alls_map = {}
    rem_map = {}
    bests = {}
    bests[N] = (u_alls_seq[0], range(N))
    for nb in range(2,N):
        bests[nb] = (float("Inf"), [])
        for tidx in itertools.combinations(range(N), nb):
            idx = list(tidx)
            u_alls_map[tidx] = (((len(idx)+2.0)/(len(idx)+1)**2)*(len(idx)*np.sum(ss[idx]**2)-np.sum(np.outer(ss[idx], ss[idx])))) 
            min_s = np.argmax(len(idx)*ss[idx]**2-2*ss[idx]*np.sum(ss[idx]))
            if u_alls_map[tidx] < bests[nb][0]:
                bests[nb] = (u_alls_map[tidx], [tidx])
            elif u_alls_map[tidx] == bests[nb][0]:
                bests[nb] = (bests[nb][0], bests[nb][1]+[tidx])
            rem_map[tidx] = idx.pop(min_s)

    for nb in range(3,N):
        if len(set(bests[nb-1][1][0])-set(bests[nb][1][0])) > 0:
            print "Not subset", bests[nb-1], bests[nb] 
    # for (k,v) in u_alls_map.items():
    #     if len(k)+1 in bests and bests[len(k)+1][0] < v:
    #         print v, k, bests[len(k)+1][0], bests[len(k)+1][1]
   
    
if __name__ == '__main__':
    # ring_exp()
    # line_exp()
    random_exp()
    # supp_line_exp()
    # tri_exp()
