import re, glob, datetime, sys, os.path
import numpy
import pdb
from experimental import *


##### COMPUTE THE NETWORK STATS (Nb edges/vertices) AND PRODUCE TEX TABLE FOR REPORTING    
xps_rep = "../latex/"
root_rep = "../data/"

filename_out = xps_rep+ "stats_networks.tex" #_"+ datetime.datetime.now().strftime('%Y%m%d%H%M%S') + ".txt"
filename_det = xps_rep+ "stats_networks_detailed.tex" #_"+ datetime.datetime.now().strftime('%Y%m%d%H%M%S') + ".txt"

groups = {"\\DBLPed{}":
          [(root_rep+"dblp/E2", "DBLP E2", "wegonet_*2-0.3.graph", "^wegonet_(?P<dataset>.*)2-0.3")],
          "\\DBLPcf{}":
          [(root_rep+"dblp/C", "DBLP L", "author-author-wpapers.*_all.txt", "^author-author-wpapers.(?P<dataset>.*)_all")],
          "\\Fb{}":
          [(root_rep+"facebook", "Facebook", "[0-9]*.edges", "^(?P<dataset>[0-9]*)")],
          "\\GPlusSM{}": 
          [(root_rep+"gplus/S", "G+ S", "[0-9]*.edges", "^(?P<dataset>[0-9]*)"),
           (root_rep+"gplus/M", "G+ M", "[0-9]*.edges", "^(?P<dataset>[0-9]*)")] ,           
          "\\GPlusXL{}": 
          [(root_rep+"gplus/L", "G+ L", "[0-9]*.edges", "^(?P<dataset>[0-9]*)"),
           (root_rep+"gplus/X", "G+ X", "[0-9]*.edges", "^(?P<dataset>[0-9]*)")]# ,
          # "XL":
          # [(root_rep+"snap", "LSN", "*.edges", "^(?P<dataset>.*).edges")]
          }

MAP_DT = {"\\GPlus{}": "G+",
    "\\Fb{}": "FB",
    "\\DBLP{}": "DBLP",
    "\\DBLPed{}": "DBLP.E2",
    "\\DBLPcf{}": "DBLP.C",
    "\\IMDB{}": "IMDB",
    "\\GPlusSM{}": "G+.S",
    "\\GPlusXL{}": "G+.L",
    "\\BKGW{}": "BKGW",
    "\\GW{}": "GW",
    "\\cXL{}": "XL"}

DT_DEF="%% Data names"
for k, v in MAP_DT.items():
    if re.match("\\\\.*\{\}", k):
        DT_DEF += "\n\\newcommand{%s}{\\texttt{%s}}" % (k.strip("{}"),v)

    
CENTERS = dict([(center.replace(" ",""), center) for center in ["Ben Shneiderman", "Erik D. Demaine", "Christos H. Papadimitriou", "Christos Faloutsos", "Heikki Mannila", "Jon M. Kleinberg", "Gerhard Weikum", "Kurt Mehlhorn", "Serge Abiteboul", "George Karypis", "Philip S. Yu", "Jiawei Han", "Martin Ester"]])


def readData(graph_file, opiext="opinions", nbo=1):
    tmp = re.search("wegonet_(?P<center>[^2]*)2", graph_file)
    if tmp is not None:
        center = CENTERS.get(tmp.group("center"), None)
    else:
        center = None
    graph = readGraph(graph_file, center=center)
    graph = largestComponent(graph)
 
    trans_graph, map_iton, map_ntoi = translate_data(graph)
    nopis = numpy.zeros((len(map_ntoi), nbo))
    # oplname = ".".join(graph_file.split(".")[:-1]+[opiext])
    # parts, dict_auts = readOpis(oplname)
    # for nn,ii in dict_auts.items():
    #     if nn in map_ntoi:
    #         nopis[map_ntoi[nn]] = parts[:nbo, ii]
    return trans_graph, nopis

def readDataNum(graph_file, opiext="opinions", nbo=1):
    graph = readGraphNum(graph_file)
    oplname = ".".join(graph_file.split(".")[:-1]+[opiext])
    nopis = readOpisNum(oplname)
    return graph, nopis


PREAMBLE_DOC="""\\documentclass[a4paper,10pt]{article}
\\usepackage[top=1cm, bottom=1cm, left=1cm, right=1cm]{geometry}
\\usepackage{booktabs}
\\begin{document}
"""

PREAMBLE_TBL="""
\\begin{table}[t]
\\centering
\\caption{%s statistics.}
\\label{tab:stats_%s}
\\footnotesize
\\begin{tabular}{%s}
\\toprule
"""

POSTAMBLE_TBL="""
\\bottomrule
\\end{tabular}
\\end{table}
"""

POSTAMBLE_DOC="""
\\end{document}
"""

fo = open(filename_out, "w")
# fo = sys.stdout
fo.write(PREAMBLE_DOC)
fo.write(DT_DEF)
fo.write(PREAMBLE_TBL % ("Data", "data", "@{\\hspace*{0.95em}}".join(["l", "r"]+ 13*["r"])))
fo.write("Collection & $\#$G & & \multicolumn{3}{c}{$|E|$} & & \multicolumn{3}{c}{$|V|$}& & \multicolumn{3}{c}{$\delta$} \\\\\n")
fo.write(" \\cline{4-6}  \\cline{8-10} \\cline{12-14}\n")
fo.write(" & & & min & max & avg & &  min & max & avg & & min & max & avg & \\\\\n\midrule\n")

fdet = open(filename_det, "w")
# fdet = sys.stdout
fdet.write(PREAMBLE_DOC)
fdet.write(DT_DEF)

for gg, series in groups.items():
    fdet.write(PREAMBLE_TBL % (gg, MAP_DT.get(gg, gg).replace(" ","-"), "@{\\hspace*{0.95em}}".join(["l", "r","r","r"])))
    fdet.write("Dataset & $|E|$ & $|V|$ & $\delta$ \\\\\n\midrule\n")

    countl = 0
    collect = {"datasets": [], "E": [], "V": [], "d": []}    
    for si, (serie, serie_t, pattt, patte) in enumerate(series):
        fnames = glob.glob(serie+"/"+pattt)
        fnames.sort(key=lambda x: os.path.getsize(x))
        if len(fnames) > 0 :
            if si > 0:
                fdet.write("[0.5em]") 
            fdet.write("\multicolumn{4}{c}{%s} \\\\\n" % serie_t)

        for gname in fnames:
            tmp = re.match(patte, gname[len(serie+"/"):])
            if tmp is None or ("101997124338642780860" in gname) or ("102778563580121606331" in gname) or ("100962871525684315897" in gname) or ("104226133029319075907" in gname):
                continue
            # graph, nopis = readDataNum(gname, "uopinions", 1)
            graph, nopis = readData(gname) #, "ropinions", 1)
            op_sqdiffs = numpy.vstack([(nopis[nA]-nopis[nBs.keys()])**2 for nA, nBs in graph.items()])

            collect["datasets"].append(tmp.group("dataset"))
            collect["V"].append(len(graph))
            collect["E"].append(len(op_sqdiffs)/2.)
            collect["d"].append(collect["E"][-1]/collect["V"][-1])
            fdet.write("%s & $%d$ & $%d$ & $%.2f$ \\\\\n" %
                       (collect["datasets"][-1], collect["E"][-1],
                        collect["V"][-1], collect["d"][-1]))

        print serie_t, len(collect["datasets"])
    if len(collect["datasets"]) > 0:
        fo.write("%s & $%d$ & & $%d$ & $%d$ & $%.2f$ & & $%d$ & $%d$ & $%.2f$ & & $%.2f$ & $%.2f$ & $%.2f$ & \\\\\n"
                 % (gg, len(collect["datasets"]),
                    numpy.min(collect["E"]), numpy.max(collect["E"]), numpy.mean(collect["E"]),
                    numpy.min(collect["V"]), numpy.max(collect["V"]), numpy.mean(collect["V"]),
                    numpy.min(collect["d"]), numpy.max(collect["d"]), numpy.mean(collect["d"])))

    fdet.write(POSTAMBLE_TBL)
fdet.write(POSTAMBLE_DOC)
fdet.close()

fo.write(POSTAMBLE_TBL)
fo.write(POSTAMBLE_DOC)
fo.close()
